# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time

import eventlet
from oslo_log import log as logging
from oslo_serialization import jsonutils
import paramiko

from tacker.common import cmd_executer
from tacker.extensions import vnfm

LOG = logging.getLogger(__name__)
HELM_CMD_TIMEOUT = 30
HELM_INSTALL_TIMEOUT = 120
HELM_UPGRADE_TIMEOUT = 120
VALUE_SPLIT_CHARACTER = '.'
TRANSPORT_RETRIES = 2
TRANSPORT_WAIT = 15


class HelmClient():
    """Helm client for hosting containerized vnfs"""

    def __init__(self, ip, username, password):
        self.host_ip = ip
        self.username = username
        self.password = password
        self.commander = cmd_executer.RemoteCommandExecutor(
            user=username,
            password=password,
            host=ip,
            timeout=HELM_CMD_TIMEOUT)

    def _execute_command(self, ssh_command, timeout=HELM_CMD_TIMEOUT, retry=0):
        eventlet.monkey_patch()
        while retry >= 0:
            try:
                with eventlet.Timeout(timeout, True):
                    result = self.commander.execute_command(
                        ssh_command, input_data=None)
                    break
            except eventlet.timeout.Timeout:
                error_message = ('It is time out, When execute command: '
                                 f'{ssh_command}.')
                LOG.debug(error_message)
                retry -= 1
                if retry < 0:
                    self.close_session()
                    LOG.error(error_message)
                    raise vnfm.HelmClientOtherError(
                        error_message=error_message)
                time.sleep(30)
        if result.get_return_code():
            self.close_session()
            err = result.get_stderr()
            LOG.error(err)
            raise vnfm.HelmClientRemoteCommandError(message=err)
        return result.get_stdout()

    def add_repository(self, repo_name, repo_url):
        # execute helm repo add command
        ssh_command = f"helm repo add {repo_name} {repo_url}"
        self._execute_command(ssh_command)

    def remove_repository(self, repo_name):
        # execute helm repo remove command
        ssh_command = f"helm repo remove {repo_name}"
        self._execute_command(ssh_command)

    def _transport_helmchart(self, source_path, target_path):
        # transfer helm chart file
        retry = TRANSPORT_RETRIES
        while retry > 0:
            try:
                connect = paramiko.Transport(self.host_ip, 22)
                connect.connect(username=self.username, password=self.password)
                sftp = paramiko.SFTPClient.from_transport(connect)
                # put helm chart file
                sftp.put(source_path, target_path)
                connect.close()
                return
            except paramiko.SSHException as e:
                LOG.debug(e)
                retry -= 1
                if retry == 0:
                    self.close_session()
                    LOG.error(e)
                    raise paramiko.SSHException()
                time.sleep(TRANSPORT_WAIT)

    def put_helmchart(self, source_path, target_dir):
        # create helm chart directory and change permission
        ssh_command = (f"if [ ! -d {target_dir} ]; then "
                       f"`sudo mkdir -p {target_dir}; "
                       f"sudo chown -R {self.username} {target_dir};`; fi")
        self._execute_command(ssh_command)
        # get helm chart name and target path
        chartfile_name = source_path[source_path.rfind(os.sep) + 1:]
        target_path = os.path.join(target_dir, chartfile_name)
        # transport helm chart file
        self._transport_helmchart(source_path, target_path)
        # decompress helm chart file
        ssh_command = f"tar -zxf {target_path} -C {target_dir}"
        self._execute_command(ssh_command)

    def delete_helmchart(self, target_path):
        # delete helm chart folder
        ssh_command = f"sudo rm -rf {target_path}"
        self._execute_command(ssh_command)

    def install(self, release_name, chart_name, namespace, parameters):
        # execute helm install command
        ssh_command = f"helm install {release_name} {chart_name}"
        if namespace:
            ssh_command = f"{ssh_command} --namespace {namespace}"
        if parameters:
            for param in parameters:
                ssh_command = f"{ssh_command} --set {param}"
        self._execute_command(ssh_command, timeout=HELM_INSTALL_TIMEOUT)

    def uninstall(self, release_name, namespace):
        # execute helm uninstall command
        ssh_command = f"helm uninstall {release_name}"
        if namespace:
            ssh_command = f"{ssh_command} --namespace {namespace}"
        self._execute_command(ssh_command, timeout=HELM_INSTALL_TIMEOUT)

    def get_manifest(self, release_name, namespace):
        # execute helm get manifest command
        ssh_command = f"helm get manifest {release_name}"
        if namespace:
            ssh_command = f"{ssh_command} --namespace {namespace}"
        result = self._execute_command(ssh_command)
        # convert manifest to text format
        mf_content = ''.join(result)
        return mf_content

    def _get_values(self, release_name, namespace):
        # execute helm get values command
        ssh_command = f"helm get values {release_name} --all --output json"
        if namespace:
            ssh_command = f"{ssh_command} --namespace {namespace}"
        result = self._execute_command(ssh_command)
        # get values (json format) from result and convert to dict
        values = jsonutils.loads(result[0])
        return values

    def get_value(self, release_name, namespace, value):
        res = self._get_values(release_name, namespace)
        # get specified value (loop for nested value: e.g. "foo.bar")
        for val in value.split(VALUE_SPLIT_CHARACTER):
            if isinstance(res, dict):
                res = res.get(val)
            else:
                # if not of type dict, target value is not found
                res = None
            if res is None:
                self.close_session()
                LOG.error(f"{value} is not found in retrieved values.")
                raise vnfm.HelmClientMissingParamsError(value=value)
        return res

    def upgrade_values(self, release_name, chart_name, namespace, parameters):
        # execute helm upgrade command
        ssh_command = (f"helm upgrade {release_name} {chart_name}"
                       " --reuse-values")
        if namespace:
            ssh_command = f"{ssh_command} --namespace {namespace}"
        for param_key, param_val in parameters.items():
            ssh_command = f"{ssh_command} --set {param_key}={param_val}"
        self._execute_command(ssh_command, timeout=HELM_UPGRADE_TIMEOUT)

    def close_session(self):
        self.commander.close_session()



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add description to vnf

Revision ID: 13c0e0661015
Revises: 4c31092895b8
Create Date: 2015-05-18 18:47:22.180962

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '13c0e0661015'
down_revision = '4c31092895b8'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.add_column('devices',
                  sa.Column('description', sa.String(255),
                            nullable=True, server_default=''))



#    Copyright 2019 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker import context
from tacker.objects import vnf_package
from tacker.objects import vnf_package_vnfd
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnfPackageVnfd(SqlTestCase):

    def setUp(self):
        super(TestVnfPackageVnfd, self).setUp()
        self.context = context.get_admin_context()

    def test_vnf_package_vnfd_create(self):
        vnf_pack = vnf_package.VnfPackage(context=self.context,
                                          **fakes.vnf_package_data)
        vnf_pack.create()
        vnf_pack_vnfd_data = fakes.vnf_pack_vnfd_data(vnf_pack.id)
        vnf_pack_vnfd_data.update({'id': uuidsentinel.id})
        result = vnf_package_vnfd._vnf_package_vnfd_create(
            self.context, vnf_pack_vnfd_data)
        self.assertTrue(result.id)
        self.assertEqual('test_provider', result.vnf_provider)

    def test_vnf_package_vnfd_get_by_id(self):
        vnf_pack = vnf_package.VnfPackage(context=self.context,
                                          **fakes.vnf_package_data)
        vnf_pack.create()

        vnf_pack_vnfd_obj = vnf_package_vnfd.VnfPackageVnfd(
            context=self.context, **fakes.vnf_pack_vnfd_data(vnf_pack.id))
        vnf_pack_vnfd_obj.create()
        result = vnf_package_vnfd._vnf_package_vnfd_get_by_id(
            self.context, vnf_pack_vnfd_obj.vnfd_id)
        self.assertEqual(vnf_pack_vnfd_obj.id, result.id)
        self.assertEqual(vnf_pack_vnfd_obj.vnf_provider, result.vnf_provider)



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

import json

from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.common import log
from tacker.extensions import vnfm
from tacker.tosca import utils as toscautils
from tacker.vnfm.infra_drivers.kubernetes.k8s import tosca_kube_object

from toscaparser.functions import GetInput
from toscaparser import tosca_template
import toscaparser.utils.yamlparser

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

YAML_LOADER = toscaparser.utils.yamlparser.load_yaml
SCALING = 'tosca.policies.Scaling'
TACKER_CP = 'tosca.nodes.nfv.CP.Tacker'
TACKER_VL = 'tosca.nodes.nfv.VL'
COLON_CHARACTER = ':'
WHITE_SPACE_CHARACTER = ' '
NON_WHITE_SPACE_CHARACTER = ''
TOSCA_LINKS_TO = 'tosca.relationships.network.LinksTo'
TOSCA_BINDS_TO = 'tosca.relationships.network.BindsTo'

ALLOWED_KUBERNETES_OBJECT_PROPS = ('namespace', 'mapping_ports', 'labels',
                                   'vnfcs', 'service_type',
                                   'mgmt_driver', 'config_drive')
ALLOWED_CONTAINER_OBJECT_PROPS = ('num_cpus', 'mem_size', 'image', 'config',
                                  'command', 'args', 'ports')
ALLOWED_SCALING_OBJECT_PROPS = ('min_instances', 'max_instances',
                                'target_cpu_utilization_percentage')

SCALAR_UNIT_DICT = {'B': 1, 'kB': 1000, 'KiB': 1024, 'MB': 1000000,
                    'MiB': 1048576, 'GB': 1000000000,
                    'GiB': 1073741824, 'TB': 1000000000000,
                    'TiB': 1099511627776}


class Parser(object):
    """Convert TOSCA template to Tosca Kube object"""

    def __init__(self, vnfd_dict):
        self.vnfd_dict = vnfd_dict

    def loader(self):
        """Load TOSCA template and start parsing"""

        try:
            parserd_params = None
            toscautils.updateimports(self.vnfd_dict)

            tosca = tosca_template.\
                ToscaTemplate(parsed_params=parserd_params,
                              a_file=False,
                              yaml_dict_tpl=self.vnfd_dict)
        except Exception as e:
            LOG.debug("tosca-parser error: %s", str(e))
            raise vnfm.ToscaParserFailed(error_msg_details=str(e))

        # Initiate a list tosca_kube_object which are defined from VDU
        tosca_kube_objects = []
        vdus = toscautils.findvdus(tosca)

        for node_template in vdus:
            vdu_name = node_template.name
            tosca_kube_obj = self.tosca_to_kube_mapping(node_template)

        # Find network name in which VDU is attached
            network_names = self.find_networks(tosca, vdu_name)
            if network_names:
                annotations_pad = \
                    json.dumps([{"name": "%s" % net}
                               for net in network_names])
                tosca_kube_obj.annotations =\
                    {'k8s.v1.cni.cncf.io/networks': annotations_pad}

            # If connection_point is True, Tacker will manage its service ip
            tosca_kube_obj.mgmt_connection_point = \
                self.check_mgmt_cp(tosca, vdu_name)

            # Find scaling policy that is used for this VDU, different to
            # VM-based VNF, there are no alarm policies.
            tosca_kube_obj.scaling_object = \
                self.get_scaling_policy(tosca, vdu_name)
            tosca_kube_objects.append(tosca_kube_obj)

        return tosca_kube_objects

    @log.log
    def tosca_to_kube_mapping(self, node_template):
        """Map TOSCA template to ToscaKubeObject properties"""
        tosca_props = self.get_properties(node_template)
        self.check_unsupported_key(tosca_props,
                                   ALLOWED_KUBERNETES_OBJECT_PROPS)
        tosca_kube_obj = tosca_kube_object.ToscaKubeObject()

        # tosca_kube_obj name is used for tracking Kubernetes resources
        service_name = 'svc-' + node_template.name + '-' + \
                       uuidutils.generate_uuid()
        tosca_kube_obj.name = service_name[:15]
        tosca_kube_obj.namespace = tosca_props.get('namespace')
        tosca_kube_obj.mapping_ports = tosca_props.get('mapping_ports')
        tosca_kube_obj.labels = tosca_props.get('labels')

        # Find config properties of VNFComponents in each VDU node
        vnfc_config_props = tosca_props.get('vnfcs')
        container_objects = self.vnfc_configurable_to_container_mapping(
            vnfc_config_props)
        tosca_kube_obj.containers = container_objects

        # In labels, we define which type of Service VNF will be deployed
        service_labels = dict()
        if tosca_kube_obj.labels:
            for label in tosca_kube_obj.labels:
                label = label.replace(
                    WHITE_SPACE_CHARACTER,
                    NON_WHITE_SPACE_CHARACTER)
                labels = label.split(COLON_CHARACTER)
                service_labels.update({labels[0]: labels[1]})
        tosca_kube_obj.labels = service_labels
        tosca_kube_obj.service_type = tosca_props.get('service_type')
        return tosca_kube_obj

    @log.log
    def vnfc_configurable_to_container_mapping(self, vnfc_config_properties):
        """Map VnfcConfigurableProperties to list of containers"""
        containers = list()
        for container_name, container_props in vnfc_config_properties.items():
            container = tosca_kube_object.Container()
            container.name = container_name
            self.check_unsupported_key(container_props,
                                       ALLOWED_CONTAINER_OBJECT_PROPS)
            container.num_cpus = container_props.get('num_cpus')
            memory_size = container_props.get('mem_size')
            container.mem_size = self.process_memory(memory_size)
            container.image = container_props.get('image')
            container.config = container_props.get('config')
            container.command = container_props.get('command')
            container.args = container_props.get('args')
            container.ports = container_props.get('ports')
            containers.append(container)
        return containers

    @log.log
    def process_memory(self, mem_value):
        """Translate memory size with unit to a number of byte memory"""
        # Memory size has the pattern e.g. 512 MB, 1024 MB or 1 GB
        parser_memory = mem_value.split(WHITE_SPACE_CHARACTER)
        memory_value = parser_memory[0]
        memory_unit = parser_memory[1]
        memory_real_value = 0

        # Translate memory's byte size based on SCALAR_UNIT_DICT
        if memory_unit in SCALAR_UNIT_DICT:
            memory_real_value = \
                int(memory_value) * SCALAR_UNIT_DICT[memory_unit]
        return memory_real_value

    @log.log
    def get_scaling_policy(self, tosca, vdu_name):
        """Find scaling policy which is used for VDU"""
        if len(tosca.policies) == 0:
            scaling_obj = None
        else:
            count = 0
            scaling_obj = tosca_kube_object.ScalingObject()
            for policy in tosca.policies:
                if policy.type_definition.is_derived_from(SCALING) \
                        and vdu_name in policy.targets:
                    count = count + 1
                    policy_props = policy.properties
                    # NOTE(ueha): check_unsupported_key() is commented out to
                    #             resolve vnf create error due to required
                    #             parameters of policies.
                    # self.check_unsupported_key(policy_props,
                    #                            ALLOWED_SCALING_OBJECT_PROPS)
                    scaling_obj.scaling_name = policy.name
                    scaling_obj.target_cpu_utilization_percentage = \
                        policy_props.get(
                            'target_cpu_utilization_percentage')
                    scaling_obj.min_replicas = \
                        policy_props.get('min_instances')
                    scaling_obj.max_replicas = \
                        policy_props.get('max_instances')

            if count > 1:
                # Because in Kubernetes environment, we can attach only one
                # scaling policy to Deployment. If user provides more than one
                # policy this error will happen when count > 1
                LOG.debug("Tacker only support one scaling policy per VDU")
                raise vnfm.InvalidKubernetesScalingPolicyNumber

        return scaling_obj

    @log.log
    def find_networks(self, tosca, vdu_name):
        """Find networks which VDU is attached based on vdu_name."""
        networks = []
        network_names = []
        for node_template in tosca.nodetemplates:
            if node_template.type_definition.is_derived_from(TACKER_CP):
                match = False
                links_to = None
                binds_to = None
                for rel, node in node_template.relationships.items():
                    if not links_to and rel.is_derived_from(TOSCA_LINKS_TO):
                        links_to = node
                    elif not binds_to and rel.is_derived_from(TOSCA_BINDS_TO):
                        binds_to = node
                        if binds_to.name == vdu_name:
                            match = True
                if match:
                    networks.append(links_to.name)

        for node_template in tosca.nodetemplates:
            if node_template.type_definition.is_derived_from(TACKER_VL):
                tosca_props = self.get_properties(node_template)
                if node_template.name in networks:
                    for key, value in tosca_props.items():
                        if key == 'network_name':
                            network_names.append(value)
        if network_names:
            return network_names

    @log.log
    def check_mgmt_cp(self, tosca, vdu_name):
        """Check if management for connection point is enabled"""
        mgmt_connection_point = False
        for nt in tosca.nodetemplates:
            if nt.type_definition.is_derived_from(TACKER_CP):
                mgmt = nt.get_property_value('management') or None
                if mgmt:
                    vdu = None
                    for rel, node in nt.relationships.items():
                        if rel.is_derived_from(TOSCA_BINDS_TO):
                            vdu = node.name
                            break
                    if vdu == vdu_name:
                        mgmt_connection_point = True
        LOG.debug('mgmt_connection_point: %s', mgmt_connection_point)
        return mgmt_connection_point

    @log.log
    def get_properties(self, node_template):
        """Return a list of property node template objects."""
        tosca_props = {}
        for prop in node_template.get_properties_objects():
            if isinstance(prop.value, GetInput):
                tosca_props[prop.name] = {'get_param': prop.value.input_name}
            else:
                tosca_props[prop.name] = prop.value
        return tosca_props

    def check_unsupported_key(self, input_values, support_key):
        """collect all unsupported keys"""
        found_keys = []
        for key in input_values:
            if key not in support_key:
                found_keys.append(key)
        if len(found_keys) > 0:
            raise vnfm.InvalidKubernetesInputParameter(found_keys=found_keys)



# Copyright 2016 Red Hat Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.services import service_base


class VNFFGPluginBase(service_base.NFVPluginBase, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_vnffgd(self, context, vnffgd):
        pass

    @abc.abstractmethod
    def delete_vnffgd(self, context, vnffgd_id):
        pass

    @abc.abstractmethod
    def get_vnffgd(self, context, vnffgd_id, fields=None):
        pass

    @abc.abstractmethod
    def get_vnffgds(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def create_vnffg(self, context, vnffg):
        pass

    @abc.abstractmethod
    def get_vnffgs(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_vnffg(self, context, vnffg_id, fields=None):
        pass

    @abc.abstractmethod
    def update_vnffg(self, context, vnffg_id, vnffg):
        pass

    @abc.abstractmethod
    def delete_vnffg(self, context, vnffg_id):
        pass

    @abc.abstractmethod
    def get_nfp(self, context, nfp_id, fields=None):
        pass

    @abc.abstractmethod
    def get_nfps(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_sfcs(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_sfc(self, context, sfc_id, fields=None):
        pass

    @abc.abstractmethod
    def get_classifiers(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_classifier(self, context, classifier_id, fields=None):
        pass



# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

import stevedore.named

from tacker._i18n import _

LOG = logging.getLogger(__name__)


class DriverManager(object):
    def __init__(self, namespace, driver_list, **kwargs):
        super(DriverManager, self).__init__()
        manager = stevedore.named.NamedExtensionManager(
            namespace, driver_list, invoke_on_load=True, **kwargs)

        drivers = {}
        for ext in manager:
            type_ = ext.obj.get_type()
            if type_ in drivers:
                msg = _("driver '%(new_driver)s' ignored because "
                        "driver '%(old_driver)s' is already "
                        "registered for driver '%(type)s'") % {
                            'new_driver': ext.name,
                            'old_driver': drivers[type].name,
                            'type': type_}
                LOG.error(msg)
                raise SystemExit(msg)
            drivers[type_] = ext
        self._drivers = dict((type_, ext.obj)
                             for (type_, ext) in drivers.items())
        LOG.info("Registered drivers from %(namespace)s: %(keys)s",
                 {'namespace': namespace, 'keys': self._drivers.keys()})

    @staticmethod
    def _driver_name(driver):
        return driver.__module__ + '.' + driver.__class__.__name__

    def register(self, type_, driver):
        if type_ in self._drivers:
            new_driver = self._driver_name(driver)
            old_driver = self._driver_name(self._drivers[type_])
            msg = _("can't load driver '%(new_driver)s' because "
                    "driver '%(old_driver)s' is already "
                    "registered for driver '%(type)s'") % {
                        'new_driver': new_driver,
                        'old_driver': old_driver,
                        'type': type_}
            LOG.error(msg)
            raise SystemExit(msg)
        self._drivers[type_] = driver

    def invoke(self, type_, method_name, **kwargs):
        driver = self._drivers[type_]
        return getattr(driver, method_name)(**kwargs)

    def __getitem__(self, type_):
        return self._drivers[type_]

    def __contains__(self, type_):
        return type_ in self._drivers



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import copy

from tacker.vnfm.lcm_user_data.abstract_user_data import AbstractUserData
import tacker.vnfm.lcm_user_data.utils as UserDataUtil


class SampleUserData(AbstractUserData):

    @staticmethod
    def instantiate(base_hot_dict=None,
                    vnfd_dict=None,
                    inst_req_info=None,
                    grant_info=None):
        api_param = UserDataUtil.get_diff_base_hot_param_from_api(
            base_hot_dict, inst_req_info)
        initial_param_dict = \
            UserDataUtil.create_initial_param_server_port_dict(
                base_hot_dict)
        vdu_flavor_dict = \
            UserDataUtil.create_vdu_flavor_capability_name_dict(vnfd_dict)
        vdu_image_dict = UserDataUtil.create_sw_image_dict(vnfd_dict)
        cpd_vl_dict = UserDataUtil.create_network_dict(
            inst_req_info, initial_param_dict)
        final_param_dict = UserDataUtil.create_final_param_dict(
            initial_param_dict, vdu_flavor_dict, vdu_image_dict, cpd_vl_dict)

        desired_capacity_dict = UserDataUtil.create_desired_capacity_dict(
            base_hot_dict, vnfd_dict, inst_req_info)
        final_param_nfv_dict = final_param_dict.get('nfv')
        if isinstance(final_param_nfv_dict, dict):
            final_param_nfv_dict.update(
                {'desired_capacity': desired_capacity_dict})

        return {**final_param_dict, **api_param}

    def heal(base_hot_dict=None,
             vnfd_dict=None,
             heal_vnf_request=None,
             vnf_instances=None,
             inst_vnf_info=None,
             param=None, vnfc_resource_info=None):

        initial_param_dict = \
            UserDataUtil.create_initial_param_server_port_dict(
                base_hot_dict)
        vdu_flavor_dict = \
            UserDataUtil.create_vdu_flavor_capability_name_dict(vnfd_dict)
        vdu_image_dict = UserDataUtil.create_sw_image_dict(vnfd_dict)
        final_param_dict = UserDataUtil.create_final_param_dict(
            initial_param_dict, vdu_flavor_dict, vdu_image_dict, {})
        for vnfc_resource in vnfc_resource_info:
            if final_param_dict.get('nfv').get('VDU').get(
                vnfc_resource.vdu_id) \
                    and param['nfv']['VDU'].get(vnfc_resource.vdu_id):
                param['nfv']['VDU'].update({
                    vnfc_resource.vdu_id:
                        final_param_dict.get('nfv').get('VDU').get(
                            vnfc_resource.vdu_id)
                })

        desired_capacity_dict = UserDataUtil.get_desired_capacity_dict(
            base_hot_dict, vnfd_dict, inst_vnf_info)
        param_nfv_dict = param.get('nfv')
        if isinstance(param_nfv_dict, dict):
            param_nfv_dict.update(
                {'desired_capacity': desired_capacity_dict})

        updated_stack_param = copy.deepcopy(param)
        return {**updated_stack_param}

    def scale(base_hot_dict=None,
             vnfd_dict=None,
             scale_vnf_request=None,
             vnf_instances=None,
             inst_vnf_info=None,
             param=None, resource_number=None):

        desired_capacity_dict = UserDataUtil.get_desired_capacity_dict(
            base_hot_dict, vnfd_dict, inst_vnf_info)
        param_nfv_dict = param.get('nfv')
        if isinstance(param_nfv_dict, dict):
            param_nfv_dict.update(
                {'desired_capacity': desired_capacity_dict})
            param_nfv_dict['desired_capacity'].update(
                {scale_vnf_request.aspect_id: resource_number})

        updated_stack_param = copy.deepcopy(param)
        return {**updated_stack_param}



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add placement table

Revision ID: 2c5211036579
Revises: ee98bbc0789d
Create Date: 2020-09-11 20:47:46.345771

"""
# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '2c5211036579'
down_revision = 'ee98bbc0789d'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import Boolean

from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.create_table(
        'placement_constraint',
        sa.Column('id', types.Uuid(length=36), nullable=False),
        sa.Column('vnf_instance_id', types.Uuid(length=36), nullable=False),
        sa.Column('affinity_or_anti_affinity',
                  sa.String(length=255), nullable=False),
        sa.Column('scope',  sa.String(length=255), nullable=False),
        sa.Column('server_group_name',  sa.String(length=255), nullable=False),
        sa.Column('resource', sa.JSON(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('deleted', Boolean, default=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB'
    )



# Copyright (c) 2019 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from tacker.common import exceptions
from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnfDeploymentFlavour(SqlTestCase):

    def setUp(self):
        super(TestVnfDeploymentFlavour, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_package = self._create_vnf_package()
        self.vnf_deployment_flavour = self._create_vnf_deployment_flavour()

    def _create_vnf_package(self):
        vnfpkgm = objects.VnfPackage(context=self.context,
                                     **fakes.vnf_package_data)
        vnfpkgm.create()
        return vnfpkgm

    def _create_vnf_deployment_flavour(self):
        flavour_data = fakes.vnf_deployment_flavour
        flavour_data.update({'package_uuid': self.vnf_package.id})
        vnf_deployment_flavour = objects.VnfDeploymentFlavour(
            context=self.context, **flavour_data)
        vnf_deployment_flavour.create()
        return vnf_deployment_flavour

    def test_create(self):
        flavour_data = fakes.vnf_deployment_flavour
        flavour_data.update({'package_uuid': self.vnf_package.id})
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(
            context=self.context, **flavour_data)

        vnf_deployment_flavour_obj.create()
        self.assertTrue(vnf_deployment_flavour_obj.id)

    def test_create_with_software_images(self):
        software_images = objects.VnfSoftwareImage(**fakes.software_image)
        fake_software_images = objects.VnfSoftwareImagesList(
            objects=[software_images])
        flavour_data = fakes.vnf_deployment_flavour
        flavour_data.update({'software_images': fake_software_images})
        flavour_data.update({'package_uuid': self.vnf_package.id})
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(
            context=self.context, **flavour_data)

        vnf_deployment_flavour_obj.create()
        self.assertTrue(vnf_deployment_flavour_obj.id)

    def test_get_by_id(self):
        vnf_deployment_flavour = objects.VnfDeploymentFlavour.get_by_id(
            self.context, self.vnf_deployment_flavour.id, expected_attrs=None)
        self.compare_obj(self.vnf_deployment_flavour, vnf_deployment_flavour,
                         allow_missing=['software_images',
                                        'updated_at',
                                        'deleted', 'deleted_at'])

    def test_get_by_id_with_no_existing_id(self):
        self.assertRaises(
            exceptions.VnfDeploymentFlavourNotFound,
            objects.VnfDeploymentFlavour.get_by_id, self.context,
            uuidsentinel.invalid_uuid)

    def test_create_with_id(self):
        vnf_deployment_flavour_obj = {'id': uuidsentinel.uuid}
        vnf_deployment_flavour = objects.VnfDeploymentFlavour(
            context=self.context, **vnf_deployment_flavour_obj)
        self.assertRaises(exceptions.ObjectActionError,
                          vnf_deployment_flavour.create)

    @mock.patch('tacker.objects.vnf_deployment_flavour.'
                '_destroy_vnf_deployment_flavour')
    def test_destroy(self, mock_vnf_deployment_flavour_destroy):
        self.vnf_deployment_flavour.destroy(self.context)
        mock_vnf_deployment_flavour_destroy.assert_called_with(
            self.context, self.vnf_deployment_flavour.id)

    def test_destroy_without_id(self):
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(
            context=self.context)
        self.assertRaises(exceptions.ObjectActionError,
                          vnf_deployment_flavour_obj.destroy, self.context)

    def test_attribute_with_valid_data(self):
        data = {'id': self.vnf_deployment_flavour.id}
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(
            context=self.context, **data)
        vnf_deployment_flavour_obj.obj_load_attr('flavour_id')
        self.assertEqual('simple', vnf_deployment_flavour_obj.flavour_id)

    def test_invalid_attribute(self):
        self.assertRaises(
            exceptions.ObjectActionError,
            self.vnf_deployment_flavour.obj_load_attr, 'invalid')

    def test_obj_load_attr_without_context(self):
        data = {'id': self.vnf_deployment_flavour.id}
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(**data)
        self.assertRaises(exceptions.OrphanedObjectError,
                          vnf_deployment_flavour_obj.obj_load_attr,
                          'flavour_id')

    def test_obj_load_attr_without_id_in_object(self):
        data = {'flavour_id': self.vnf_deployment_flavour.flavour_id}
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(
            context=self.context, **data)
        self.assertRaises(exceptions.ObjectActionError,
                          vnf_deployment_flavour_obj.obj_load_attr,
                          'flavour_id')

    def test_get_by_id_with_software_images(self):
        software_image = fakes.software_image
        software_image.update(
            {'flavour_uuid': self.vnf_deployment_flavour.id})

        vnf_soft_image_obj = objects.VnfSoftwareImage(
            context=self.context, **software_image)
        vnf_soft_image_obj.create()
        vnf_deployment_flavour = objects.VnfDeploymentFlavour.get_by_id(
            self.context, self.vnf_deployment_flavour.id,
            expected_attrs=['software_images'])
        self.assertEqual(1,
                         len(vnf_deployment_flavour.software_images.objects))
        self.compare_obj(vnf_soft_image_obj,
                         vnf_deployment_flavour.software_images[0])



# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import netaddr
from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker._i18n import _
from tacker.common import exceptions as n_exc


LOG = logging.getLogger(__name__)

ATTRIBUTES_TO_UPDATE = 'attributes_to_update'
ATTR_NOT_SPECIFIED = object()
# Defining a constant to avoid repeating string literal in several modules
SHARED = 'shared'

# Used by range check to indicate no limit for a bound.
UNLIMITED = None


def _verify_dict_keys(expected_keys, target_dict, strict=True):
    """Allows to verify keys in a dictionary.

    :param expected_keys: A list of keys expected to be present.
    :param target_dict: The dictionary which should be verified.
    :param strict: Specifies whether additional keys are allowed to be present.
    :return: True, if keys in the dictionary correspond to the specification.
    """
    if not isinstance(target_dict, dict):
        msg = (_("Invalid input. '%(target_dict)s' must be a dictionary "
                 "with keys: %(expected_keys)s") %
               {'target_dict': target_dict, 'expected_keys': expected_keys})
        return msg

    expected_keys = set(expected_keys)
    provided_keys = set(target_dict.keys())

    predicate = expected_keys.__eq__ if strict else expected_keys.issubset

    if not predicate(provided_keys):
        msg = (_("Validation of dictionary's keys failed. "
                 "Expected keys: %(expected_keys)s "
                 "Provided keys: %(provided_keys)s") %
               {'expected_keys': expected_keys,
                'provided_keys': provided_keys})
        return msg


def is_attr_set(attribute):
    return not (attribute is None or attribute is ATTR_NOT_SPECIFIED)


def _validate_values(data, valid_values=None):
    if data not in valid_values:
        msg = (_("'%(data)s' is not in %(valid_values)s") %
               {'data': data, 'valid_values': valid_values})
        LOG.debug(msg)
        return msg


def _validate_not_empty_string_or_none(data, max_len=None):
    if data is not None:
        return _validate_not_empty_string(data, max_len=max_len)


def _validate_not_empty_string(data, max_len=None):
    msg = _validate_string(data, max_len=max_len)
    if msg:
        return msg
    if not data.strip():
        return _("'%s' Blank strings are not permitted") % data


def _validate_string_or_none(data, max_len=None):
    if data is not None:
        return _validate_string(data, max_len=max_len)


def _validate_string(data, max_len=None):
    if not isinstance(data, str):
        msg = _("'%s' is not a valid string") % data
        LOG.debug(msg)
        return msg

    if max_len is not None and len(data) > max_len:
        msg = (_("'%(data)s' exceeds maximum length of %(max_len)s") %
               {'data': data, 'max_len': max_len})
        LOG.debug(msg)
        return msg


def _validate_boolean(data, valid_values=None):
    try:
        convert_to_boolean(data)
    except n_exc.InvalidInput:
        msg = _("'%s' is not a valid boolean value") % data
        LOG.debug(msg)
        return msg


def _validate_range(data, valid_values=None):
    """Check that integer value is within a range provided.

    Test is inclusive. Allows either limit to be ignored, to allow
    checking ranges where only the lower or upper limit matter.
    It is expected that the limits provided are valid integers or
    the value None.
    """

    min_value = valid_values[0]
    max_value = valid_values[1]
    try:
        data = int(data)
    except (ValueError, TypeError):
        msg = _("'%s' is not an integer") % data
        LOG.debug(msg)
        return msg
    if min_value is not UNLIMITED and data < min_value:
        msg = _("'%(data)s' is too small - must be at least "
                "'%(limit)d'") % {'data': data, 'limit': min_value}
        LOG.debug(msg)
        return msg
    if max_value is not UNLIMITED and data > max_value:
        msg = _("'%(data)s' is too large - must be no larger than "
                "'%(limit)d'") % {'data': data, 'limit': max_value}
        LOG.debug(msg)
        return msg


def _validate_no_whitespace(data):
    """Validates that input has no whitespace."""
    if len(data.split()) > 1:
        msg = _("'%s' contains whitespace") % data
        LOG.debug(msg)
        raise n_exc.InvalidInput(error_message=msg)
    return data


def _validate_mac_address(data, valid_values=None):
    try:
        valid_mac = netaddr.valid_mac(_validate_no_whitespace(data))
        if valid_mac is False:
            msg = _("'%s' is not a valid MAC address") % data
            LOG.debug(msg)
            return msg
    except AttributeError as ex:
        msg = _("MAC address must be string: %s") % ex
        LOG.exception(msg)
        return msg


def _validate_mac_address_or_none(data, valid_values=None):
    if data is None:
        return
    return _validate_mac_address(data, valid_values)


def _validate_ip_address(data, valid_values=None):
    try:
        netaddr.IPAddress(_validate_no_whitespace(data))
    except Exception:
        msg = _("'%s' is not a valid IP address") % data
        LOG.debug(msg)
        return msg


def _validate_ip_pools(data, valid_values=None):
    """Validate that start and end IP addresses are present.

    In addition to this the IP addresses will also be validated
    """
    if not isinstance(data, list):
        msg = _("Invalid data format for IP pool: '%s'") % data
        LOG.debug(msg)
        return msg

    expected_keys = ['start', 'end']
    for ip_pool in data:
        msg = _verify_dict_keys(expected_keys, ip_pool)
        if msg:
            LOG.debug(msg)
            return msg
        for k in expected_keys:
            msg = _validate_ip_address(ip_pool[k])
            if msg:
                LOG.debug(msg)
                return msg


def _validate_fixed_ips(data, valid_values=None):
    if not isinstance(data, list):
        msg = _("Invalid data format for fixed IP: '%s'") % data
        LOG.debug(msg)
        return msg

    ips = []
    for fixed_ip in data:
        if not isinstance(fixed_ip, dict):
            msg = _("Invalid data format for fixed IP: '%s'") % fixed_ip
            LOG.debug(msg)
            return msg
        if 'ip_address' in fixed_ip:
            # Ensure that duplicate entries are not set - just checking IP
            # suffices. Duplicate subnet_id's are legitimate.
            fixed_ip_address = fixed_ip['ip_address']
            if fixed_ip_address in ips:
                msg = _("Duplicate IP address '%s'") % fixed_ip_address
            else:
                msg = _validate_ip_address(fixed_ip_address)
            if msg:
                LOG.debug(msg)
                return msg
            ips.append(fixed_ip_address)
        if 'subnet_id' in fixed_ip:
            msg = _validate_uuid(fixed_ip['subnet_id'])
            if msg:
                LOG.debug(msg)
                return msg


def _validate_nameservers(data, valid_values=None):
    if not hasattr(data, '__iter__'):
        msg = _("Invalid data format for nameserver: '%s'") % data
        LOG.debug(msg)
        return msg

    ips = []
    for ip in data:
        msg = _validate_ip_address(ip)
        if msg:
            # This may be a hostname
            msg = _validate_regex(ip, HOSTNAME_PATTERN)
            if msg:
                msg = _("'%s' is not a valid nameserver") % ip
                LOG.debug(msg)
                return msg
        if ip in ips:
            msg = _("Duplicate nameserver '%s'") % ip
            LOG.debug(msg)
            return msg
        ips.append(ip)


def _validate_hostroutes(data, valid_values=None):
    if not isinstance(data, list):
        msg = _("Invalid data format for hostroute: '%s'") % data
        LOG.debug(msg)
        return msg

    expected_keys = ['destination', 'nexthop']
    hostroutes = []
    for hostroute in data:
        msg = _verify_dict_keys(expected_keys, hostroute)
        if msg:
            LOG.debug(msg)
            return msg
        msg = _validate_subnet(hostroute['destination'])
        if msg:
            LOG.debug(msg)
            return msg
        msg = _validate_ip_address(hostroute['nexthop'])
        if msg:
            LOG.debug(msg)
            return msg
        if hostroute in hostroutes:
            msg = _("Duplicate hostroute '%s'") % hostroute
            LOG.debug(msg)
            return msg
        hostroutes.append(hostroute)


def _validate_ip_address_or_none(data, valid_values=None):
    if data is None:
        return None
    return _validate_ip_address(data, valid_values)


def _validate_subnet(data, valid_values=None):
    msg = None
    try:
        net = netaddr.IPNetwork(_validate_no_whitespace(data))
        if '/' not in data:
            msg = _("'%(data)s' isn't a recognized IP subnet cidr,"
                    " '%(cidr)s' is recommended") % {"data": data,
                                                     "cidr": net.cidr}
        else:
            return
    except Exception:
        msg = _("'%s' is not a valid IP subnet") % data
    if msg:
        LOG.debug(msg)
    return msg


def _validate_subnet_list(data, valid_values=None):
    if not isinstance(data, list):
        msg = _("'%s' is not a list") % data
        LOG.debug(msg)
        return msg

    if len(set(data)) != len(data):
        msg = _("Duplicate items in the list: '%s'") % ', '.join(data)
        LOG.debug(msg)
        return msg

    for item in data:
        msg = _validate_subnet(item)
        if msg:
            return msg


def _validate_subnet_or_none(data, valid_values=None):
    if data is None:
        return
    return _validate_subnet(data, valid_values)


def _validate_regex(data, valid_values=None):
    try:
        if re.match(valid_values, data):
            return
    except TypeError:
        pass

    msg = _("'%s' is not a valid input") % data
    LOG.debug(msg)
    return msg


def _validate_regex_or_none(data, valid_values=None):
    if data is None:
        return
    return _validate_regex(data, valid_values)


def _validate_uuid(data, valid_values=None):
    if not uuidutils.is_uuid_like(data):
        msg = _("'%s' is not a valid UUID") % data
        LOG.debug(msg)
        return msg


def _validate_uuid_or_none(data, valid_values=None):
    if data is not None:
        return _validate_uuid(data)


def _validate_uuid_list(data, valid_values=None):
    if not isinstance(data, list):
        msg = _("'%s' is not a list") % data
        LOG.debug(msg)
        return msg

    for item in data:
        msg = _validate_uuid(item)
        if msg:
            LOG.debug(msg)
            return msg

    if len(set(data)) != len(data):
        msg = _("Duplicate items in the list: '%s'") % ', '.join(data)
        LOG.debug(msg)
        return msg


def _validate_dict_item(key, key_validator, data):
    # Find conversion function, if any, and apply it
    conv_func = key_validator.get('convert_to')
    if conv_func:
        data[key] = conv_func(data.get(key))
    # Find validator function
    # TODO(salv-orlando): Structure of dict attributes should be improved
    # to avoid iterating over items
    val_func = val_params = None
    for (k, v) in (key_validator).items():
        if k.startswith('type:'):
            # ask forgiveness, not permission
            try:
                val_func = validators[k]
            except KeyError:
                return _("Validator '%s' does not exist.") % k
            val_params = v
            break
    # Process validation
    if val_func:
        return val_func(data.get(key), val_params)


def _validate_dict(data, key_specs=None):
    if not isinstance(data, dict):
        msg = _("'%s' is not a dictionary") % data
        LOG.debug(msg)
        return msg
    # Do not perform any further validation, if no constraints are supplied
    if not key_specs:
        return

    # Check whether all required keys are present
    required_keys = [key for key, spec in (key_specs).items()
                     if spec.get('required')]

    if required_keys:
        msg = _verify_dict_keys(required_keys, data, False)
        if msg:
            LOG.debug(msg)
            return msg

    # Perform validation and conversion of all values
    # according to the specifications.
    for key, key_validator in [(k, v) for k, v in (key_specs).items()
                               if k in data]:
        msg = _validate_dict_item(key, key_validator, data)
        if msg:
            LOG.debug(msg)
            return msg


def _validate_dict_or_none(data, key_specs=None):
    if data is not None:
        return _validate_dict(data, key_specs)


def _validate_dict_or_empty(data, key_specs=None):
    if data != {}:
        return _validate_dict(data, key_specs)


def _validate_dict_or_nodata(data, key_specs=None):
    if data:
        return _validate_dict(data, key_specs)


def _validate_dict_not_empty(data, key_specs=None):
    if not isinstance(data, dict) or len(data) == 0:
        msg = _("'%s' is not a valid dictionary or it is an"
                " empty dictionary") % data
        LOG.debug(msg)
        return msg
    else:
        return _validate_dict(data, key_specs)


def _validate_non_negative(data, valid_values=None):
    try:
        data = int(data)
    except (ValueError, TypeError):
        msg = _("'%s' is not an integer") % data
        LOG.debug(msg)
        return msg

    if data < 0:
        msg = _("'%s' should be non-negative") % data
        LOG.debug(msg)
        return msg


def convert_to_boolean(data):
    if isinstance(data, str):
        val = data.lower()
        if val == "true" or val == "1":
            return True
        if val == "false" or val == "0":
            return False
    elif isinstance(data, bool):
        return data
    elif isinstance(data, int):
        if data == 0:
            return False
        elif data == 1:
            return True
    msg = _("'%s' cannot be converted to boolean") % data
    raise n_exc.InvalidInput(error_message=msg)


def convert_to_int(data):
    try:
        return int(data)
    except (ValueError, TypeError):
        msg = _("'%s' is not an integer") % data
        raise n_exc.InvalidInput(error_message=msg)


def convert_kvp_str_to_list(data):
    """Convert a value of the form 'key=value' to ['key', 'value'].

    :raises n_exc.InvalidInput: if any of the strings are malformed
                                (e.g. do not contain a key).
    """
    kvp = [x.strip() for x in data.split('=', 1)]
    if len(kvp) == 2 and kvp[0]:
        return kvp
    msg = _("'%s' is not of the form <key>=[value]") % data
    raise n_exc.InvalidInput(error_message=msg)


def convert_kvp_list_to_dict(kvp_list):
    """Convert a list of 'key=value' strings to a dict.

    :raises n_exc.InvalidInput: if any of the strings are malformed
                                (e.g. do not contain a key) or if any
                                of the keys appear more than once.
    """
    if kvp_list == ['True']:
        # No values were provided (i.e. '--flag-name')
        return {}
    kvp_map = {}
    for kvp_str in kvp_list:
        key, value = convert_kvp_str_to_list(kvp_str)
        kvp_map.setdefault(key, set())
        kvp_map[key].add(value)
    return dict((x, list(y)) for x, y in (kvp_map).items())


def convert_none_to_empty_list(value):
    return [] if value is None else value


def convert_none_to_empty_dict(value):
    return {} if value is None else value


def convert_to_list(data):
    if data is None:
        return []
    elif hasattr(data, '__iter__') and not isinstance(data, str):
        return list(data)
    else:
        return [data]


HOSTNAME_PATTERN = (r"(?=^.{1,254}$)(^(?:(?!\d+\.|-)[a-zA-Z0-9_\-]"
                    r"{1,63}(?<!-)\.?)+(?:[a-zA-Z]{2,})$)")

HEX_ELEM = '[0-9A-Fa-f]'
UUID_PATTERN = '-'.join([HEX_ELEM + '{8}', HEX_ELEM + '{4}',
                         HEX_ELEM + '{4}', HEX_ELEM + '{4}',
                         HEX_ELEM + '{12}'])
# Note: In order to ensure that the MAC address is unicast the first byte
# must be even.
MAC_PATTERN = "^%s[aceACE02468](:%s{2}){5}$" % (HEX_ELEM, HEX_ELEM)

# Dictionary that maintains a list of validation functions
validators = {'type:dict': _validate_dict,
              'type:dict_or_none': _validate_dict_or_none,
              'type:dict_or_empty': _validate_dict_or_empty,
              'type:dict_or_nodata': _validate_dict_or_nodata,
              'type:dict_not_empty': _validate_dict_not_empty,
              'type:fixed_ips': _validate_fixed_ips,
              'type:hostroutes': _validate_hostroutes,
              'type:ip_address': _validate_ip_address,
              'type:ip_address_or_none': _validate_ip_address_or_none,
              'type:ip_pools': _validate_ip_pools,
              'type:mac_address': _validate_mac_address,
              'type:mac_address_or_none': _validate_mac_address_or_none,
              'type:nameservers': _validate_nameservers,
              'type:non_negative': _validate_non_negative,
              'type:range': _validate_range,
              'type:regex': _validate_regex,
              'type:regex_or_none': _validate_regex_or_none,
              'type:string': _validate_string,
              'type:string_or_none': _validate_string_or_none,
              'type:not_empty_string': _validate_not_empty_string,
              'type:not_empty_string_or_none':
              _validate_not_empty_string_or_none,
              'type:subnet': _validate_subnet,
              'type:subnet_list': _validate_subnet_list,
              'type:subnet_or_none': _validate_subnet_or_none,
              'type:uuid': _validate_uuid,
              'type:uuid_or_none': _validate_uuid_or_none,
              'type:uuid_list': _validate_uuid_list,
              'type:values': _validate_values,
              'type:boolean': _validate_boolean}

# Define constants for base resource name

# Note: a default of ATTR_NOT_SPECIFIED indicates that an
# attribute is not required, but will be generated by the plugin
# if it is not specified.  Particularly, a value of ATTR_NOT_SPECIFIED
# is different from an attribute that has been specified with a value of
# None.  For example, if 'gateway_ip' is omitted in a request to
# create a subnet, the plugin will receive ATTR_NOT_SPECIFIED
# and the default gateway_ip will be generated.
# However, if gateway_ip is specified as None, this means that
# the subnet does not have a gateway IP.
# The following is a short reference for understanding attribute info:
# default: default value of the attribute (if missing, the attribute
# becomes mandatory.
# allow_post: the attribute can be used on POST requests.
# allow_put: the attribute can be used on PUT requests.
# validate: specifies rules for validating data in the attribute.
# convert_to: transformation to apply to the value before it is returned
# is_visible: the attribute is returned in GET responses.
# required_by_policy: the attribute is required by the policy engine and
# should therefore be filled by the API layer even if not present in
# request body.
# enforce_policy: the attribute is actively part of the policy enforcing
# mechanism, ie: there might be rules which refer to this attribute.

# Identify the attribute used by a resource to reference another resource

RESOURCE_ATTRIBUTE_MAP = {}

RESOURCE_FOREIGN_KEYS = {}

PLURALS = {'extensions': 'extension'}



# Copyright 2013 VMware, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from tacker.api import extensions
from tacker.api.v1 import base
from tacker import manager


# Attribute Map
RESOURCE_ATTRIBUTE_MAP = {
    'ext_test_resources': {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None},
               'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None},
                 'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'required_by_policy': True,
                      'validate': {'type:string': None},
                      'is_visible': True},
    }
}


class Extensionattribute(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Extension Test Resource"

    @classmethod
    def get_alias(cls):
        return "ext-obj-test"

    @classmethod
    def get_description(cls):
        return "Extension Test Resource"

    @classmethod
    def get_namespace(cls):
        return ""

    @classmethod
    def get_updated(cls):
        return "2013-02-05T10:00:00-00:00"

    def update_attributes_map(self, attributes):
        super(Extensionattribute, self).update_attributes_map(
            attributes, extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        exts = []
        plugin = manager.TackerManager.get_plugin()
        resource_name = 'ext_test_resource'
        collection_name = resource_name + "s"
        params = RESOURCE_ATTRIBUTE_MAP.get(collection_name, dict())

        controller = base.create_resource(collection_name,
                                          resource_name,
                                          plugin, params,
                                          member_actions={})

        ex = extensions.ResourceExtension(collection_name,
                                          controller,
                                          member_actions={})
        exts.append(ex)

        return exts

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


class ExtensionObjectTestPluginBase(object):

    @abc.abstractmethod
    def create_ext_test_resource(self, context, router):
        pass

    @abc.abstractmethod
    def get_ext_test_resource(self, context, id, fields=None):
        pass



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import base
from tacker.sol_refactored.objects import fields
from tacker.sol_refactored.objects.v2 import fields as v2fields


# NFV-SOL 005
# - v3.3.1 9.5.3.4 (API version: 2.1.0)
@base.TackerObjectRegistry.register
class PkgmNotificationFilterV2(base.TackerObject, base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'notificationTypes': fields.ListOfEnumField(
            valid_values=[
                'VnfPackageOnboardingNotification',
                'VnfPackageChangeNotification',
            ],
            nullable=True,
        ),
        'vnfProductsFromProviders': fields.ListOfObjectsField(
            'PkgmNotificationFilterV2_VnfProductsFromProviders',
            nullable=True),
        'vnfdId': fields.ListOfStringsField(nullable=True),
        'vnfPkgId': fields.ListOfStringsField(nullable=True),
        'operationalState': fields.Field(fields.List(
            v2fields.PackageOperationalStateType), nullable=True),
        'usageState': fields.Field(fields.List(
            v2fields.PackageUsageStateType), nullable=True),
        'vnfmInfo': fields.ListOfStringsField(nullable=True),
    }


@base.TackerObjectRegistry.register
class PkgmNotificationFilterV2_VnfProductsFromProviders(
        base.TackerObject, base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vnfProvider': fields.StringField(nullable=False),
        'vnfProducts': fields.ListOfObjectsField(
            'PkgmNotificationFilterV2_VnfProductsFromProviders_VnfProducts',
            nullable=True),
    }


@base.TackerObjectRegistry.register
class PkgmNotificationFilterV2_VnfProductsFromProviders_VnfProducts(
        base.TackerObject, base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vnfProductName': fields.StringField(nullable=False),
        'vnfVersions': fields.ListOfObjectsField(
            'PkgmNotificationFilterV2_VnfProductsFromProviders_'
            'VnfProducts_Versions',
            nullable=True),
    }


@base.TackerObjectRegistry.register
class PkgmNotificationFilterV2_VnfProductsFromProviders_VnfProducts_Versions(
        base.TackerObject, base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vnfSoftwareVersion': fields.VersionField(nullable=False),
        'vnfdVersion': fields.VersionField(nullable=False),
    }



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Utility methods for working with WSGI servers
"""
import functools

import errno
import http.client
import os
import socket
import ssl
import sys
import time

import eventlet.wsgi
# eventlet.patcher.monkey_patch(all=False, socket=True, thread=True)
from oslo_config import cfg
import tacker.conf

import oslo_i18n as i18n
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_service import systemd
from oslo_utils import encodeutils
from oslo_utils import excutils
import routes.middleware
import webob.dec
import webob.exc

from tacker._i18n import _
from tacker.common import exceptions as exception
from tacker import context


socket_opts = [
    cfg.IntOpt('backlog',
               default=4096,
               help=_("Number of backlog requests to configure "
                      "the socket with")),
    cfg.IntOpt('tcp_keepidle',
               default=600,
               help=_("Sets the value of TCP_KEEPIDLE in seconds for each "
                      "server socket. Not supported on OS X.")),
    cfg.IntOpt('retry_until_window',
               default=30,
               help=_("Number of seconds to keep retrying to listen")),
    cfg.IntOpt('max_header_line',
               default=16384,
               help=_("Max header line to accommodate large tokens")),
    cfg.BoolOpt('use_ssl',
                default=False,
                help=_('Enable SSL on the API server')),
    cfg.StrOpt('ssl_ca_file',
               help=_("CA certificate file to use to verify "
                      "connecting clients")),
    cfg.StrOpt('ssl_cert_file',
               help=_("Certificate file to use when starting "
                      "the server securely")),
    cfg.StrOpt('ssl_key_file',
               help=_("Private key file to use when starting "
                      "the server securely")),
]

CONF = tacker.conf.CONF
CONF.register_opts(socket_opts)


def config_opts():
    return [(None, socket_opts)]


LOG = logging.getLogger(__name__)


def encode_body(body):
    """Encode unicode body.

    WebOb requires to encode unicode body used to update response body.
    """
    return encodeutils.to_utf8(body)


def expected_errors(errors):
    """Decorator for Restful API methods which specifies expected exceptions.

    Specify which exceptions may occur when an API method is called. If an
    unexpected exception occurs then return a 500 instead and ask the user
    of the API to file a bug report.
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as exc:
                if isinstance(exc, webob.exc.WSGIHTTPException):
                    if isinstance(errors, int):
                        t_errors = (errors,)
                    else:
                        t_errors = errors
                    if exc.code in t_errors:
                        raise
                elif isinstance(exc, exception.Forbidden):
                    # Note(nirajsingh): Special case to handle
                    # Forbidden exceptions so every
                    # extension method does not need to wrap authorize
                    # calls. ResourceExceptionHandler silently
                    # converts NotAuthorized to HTTPForbidden
                    raise
                elif isinstance(exc, exception.ValidationError):
                    # Note(nirajsingh): Handle a validation error, which
                    # happens due to invalid API parameters, as an
                    # expected error.
                    raise
                elif isinstance(exc, exception.NotAuthorized):
                    # Handle an authorized exception, will be
                    # automatically converted to a HTTP 401.
                    raise
                elif isinstance(exc, exception.Conflict):
                    # Note(tpatil): Handle a conflict error, which
                    # happens due to resources in wrong state.
                    # ResourceExceptionHandler silently converts Conflict
                    # to HTTPConflict
                    raise

                LOG.exception("Unexpected exception in API method")
                msg = _('Unexpected API Error. Please report this at '
                    'http://bugs.launchpad.net/tacker/ and attach the Tacker '
                    'API log if possible.\n%s') % type(exc)
                raise webob.exc.HTTPInternalServerError(explanation=msg)

        return wrapped

    return decorator


class Server(object):
    """Server class to manage multiple WSGI sockets and applications."""

    def __init__(self, name, threads=1000):
        # Raise the default from 8192 to accommodate large tokens
        eventlet.wsgi.MAX_HEADER_LINE = CONF.max_header_line
        self.pool = eventlet.GreenPool(threads)
        self.name = name
        self._server = None

    def _get_socket(self, host, port, backlog):
        bind_addr = (host, port)
        # TODO(dims): eventlet's green dns/socket module does not actually
        # support IPv6 in getaddrinfo(). We need to get around this in the
        # future or monitor upstream for a fix
        try:
            info = socket.getaddrinfo(bind_addr[0],
                                      bind_addr[1],
                                      socket.AF_UNSPEC,
                                      socket.SOCK_STREAM)[0]
            family = info[0]
            bind_addr = info[-1]
        except Exception:
            LOG.exception("Unable to listen on %(host)s:%(port)s",
                          {'host': host, 'port': port})
            sys.exit(1)

        if CONF.use_ssl:
            if not os.path.exists(CONF.ssl_cert_file):
                raise RuntimeError(_("Unable to find ssl_cert_file"
                                     ": %s") % CONF.ssl_cert_file)

            # ssl_key_file is optional because the key may be embedded in the
            # certificate file
            if CONF.ssl_key_file and not os.path.exists(CONF.ssl_key_file):
                raise RuntimeError(_("Unable to find "
                                     "ssl_key_file: %s") % CONF.ssl_key_file)

            # ssl_ca_file is optional
            if CONF.ssl_ca_file and not os.path.exists(CONF.ssl_ca_file):
                raise RuntimeError(_("Unable to find ssl_ca_file"
                                     ": %s") % CONF.ssl_ca_file)

        def wrap_ssl(sock):
            ssl_kwargs = {
                'server_side': True,
                'certfile': CONF.ssl_cert_file,
                'keyfile': CONF.ssl_key_file,
                'cert_reqs': ssl.CERT_NONE,
            }

            if CONF.ssl_ca_file:
                ssl_kwargs['ca_certs'] = CONF.ssl_ca_file
                ssl_kwargs['cert_reqs'] = ssl.CERT_REQUIRED

            return ssl.wrap_socket(sock, **ssl_kwargs)

        sock = None
        retry_until = time.time() + CONF.retry_until_window
        while not sock and time.time() < retry_until:
            try:
                sock = eventlet.listen(bind_addr,
                                       backlog=backlog,
                                       family=family)
                if CONF.use_ssl:
                    sock = wrap_ssl(sock)
            except socket.error as err:
                with excutils.save_and_reraise_exception() as ctxt:
                    if err.errno == errno.EADDRINUSE:
                        ctxt.reraise = False
                        eventlet.sleep(0.1)
        if not sock:
            raise RuntimeError(_("Could not bind to %(host)s:%(port)s "
                                 "after trying for %(time)d seconds") %
                               {'host': host,
                                'port': port,
                                'time': CONF.retry_until_window})
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # sockets can hang around forever without keepalive
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        # This option isn't available in the OS X version of eventlet
        if hasattr(socket, 'TCP_KEEPIDLE'):
            sock.setsockopt(socket.IPPROTO_TCP,
                            socket.TCP_KEEPIDLE,
                            CONF.tcp_keepidle)

        return sock

    def start(self, application, port, host='0.0.0.0'):
        """Run a WSGI server with the given application."""
        self._host = host
        self._port = port
        backlog = CONF.backlog

        self._socket = self._get_socket(self._host,
                                        self._port,
                                        backlog=backlog)
        # For the case where only one process is required.
        self._server = self.pool.spawn(self._run, application,
                                       self._socket)
        systemd.notify_once()

    @property
    def host(self):
        return self._socket.getsockname()[0] if self._socket else self._host

    @property
    def port(self):
        return self._socket.getsockname()[1] if self._socket else self._port

    def stop(self):
        self._server.kill()

    def wait(self):
        """Wait until all servers have completed running."""
        try:
            self.pool.waitall()
        except KeyboardInterrupt:
            pass

    def _run(self, application, socket):
        """Start a WSGI server in a new green thread."""
        eventlet.wsgi.server(socket, application, custom_pool=self.pool,
                             log=LOG)


class Middleware(object):
    """Base WSGI middleware wrapper.

    These classes require an application to be initialized that will be called
    next.  By default the middleware will simply call its wrapped app, or you
    can override __call__ to customize its behavior.
    """

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [filter:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [filter:analytics]
            redis_host = 127.0.0.1
            paste.filter_factory = nova.api.analytics:Analytics.factory

        which would result in a call to the `Analytics` class as

            import nova.api.analytics
            analytics.Analytics(app_from_paste, redis_host='127.0.0.1')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        def _factory(app):
            return cls(app, **local_config)
        return _factory

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.

        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @webob.dec.wsgify
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)


class Request(webob.Request):

    def best_match_content_type(self):
        """Determine the most acceptable content-type.

        Based on:
            1) URI extension (.json)
            2) Content-type header
            3) Accept* headers
        """
        # First lookup http request path
        parts = self.path.rsplit('.', 1)
        if len(parts) > 1:
            _format = parts[1]
            if _format in ['json']:
                return 'application/{0}'.format(_format)

        # Then look up content header
        type_from_header = self.get_content_type()
        if type_from_header:
            return type_from_header
        ctypes = ['application/json', 'text/plain', 'application/zip']

        # Finally search in Accept-* headers
        bm = self.accept.best_match(ctypes)
        return bm or 'application/json'

    def get_content_type(self):
        allowed_types = ("application/json", "application/zip")
        if "Content-Type" not in self.headers:
            LOG.debug("Missing Content-Type")
            return None
        _type = self.content_type
        if _type in allowed_types:
            return _type
        return None

    def best_match_language(self):
        """Determines best available locale from the Accept-Language header.

        :returns: the best language match or None if the 'Accept-Language'
                  header was not available in the request.
        """
        if not self.accept_language:
            return None
        all_languages = i18n.get_available_languages('tacker')
        return self.accept_language.best_match(all_languages)

    @property
    def context(self):
        if 'tacker.context' not in self.environ:
            self.environ['tacker.context'] = context.get_admin_context()
        return self.environ['tacker.context']


class ActionDispatcher(object):
    """Maps method name to local methods through action name."""

    def dispatch(self, *args, **kwargs):
        """Find and call local method."""
        action = kwargs.pop('action', 'default')
        action_method = getattr(self, str(action), self.default)
        return action_method(*args, **kwargs)

    def default(self, data):
        raise NotImplementedError()


class DictSerializer(ActionDispatcher):
    """Default request body serialization."""

    def serialize(self, data, action='default'):
        return self.dispatch(data, action=action)

    def default(self, data):
        return ""


class JSONDictSerializer(DictSerializer):
    """Default JSON request body serialization."""

    def default(self, data):
        def sanitizer(obj):
            return str(obj)
        return encode_body(jsonutils.dump_as_bytes(data, default=sanitizer))


class ResponseHeaderSerializer(ActionDispatcher):
    """Default response headers serialization."""

    def serialize(self, response, data, action):
        self.dispatch(response, data, action=action)

    def default(self, response, data):
        response.status_int = 200


class ResponseSerializer(object):
    """Encode the necessary pieces into a response object."""

    def __init__(self, body_serializers=None, headers_serializer=None):
        self.body_serializers = {
            'application/json': JSONDictSerializer(),
            'application/zip': JSONDictSerializer()
        }
        self.body_serializers.update(body_serializers or {})

        self.headers_serializer = (headers_serializer or
                                   ResponseHeaderSerializer())

    def serialize(self, response_data, content_type, action='default'):
        """Serialize a dict into a string and wrap in a wsgi.Request object.

        :param response_data: dict produced by the Controller
        :param content_type: expected mimetype of serialized response body

        """
        response = webob.Response()
        self.serialize_headers(response, response_data, action)
        self.serialize_body(response, response_data, content_type, action)
        return response

    def serialize_headers(self, response, data, action):
        self.headers_serializer.serialize(response, data, action)

    def serialize_body(self, response, data, content_type, action):
        response.headers['Content-Type'] = content_type
        if data is not None:
            serializer = self.get_body_serializer(content_type)
            response.body = serializer.serialize(data, action)

    def get_body_serializer(self, content_type):
        try:
            return self.body_serializers[content_type]
        except (KeyError, TypeError):
            raise exception.InvalidContentType(content_type=content_type)


class TextDeserializer(ActionDispatcher):
    """Default request body deserialization."""

    def deserialize(self, datastring, action='default'):
        return self.dispatch(datastring, action=action)

    def default(self, datastring):
        return {}


class JSONDeserializer(TextDeserializer):

    def _from_json(self, datastring):
        try:
            return jsonutils.loads(datastring)
        except ValueError:
            msg = _("Cannot understand JSON")
            raise exception.MalformedRequestBody(reason=msg)

    def default(self, datastring):
        return {'body': self._from_json(datastring)}


class ZipDeserializer(ActionDispatcher):

    def deserialize(self, body_file, action='default'):
        return self.dispatch(body_file, action=action)

    def default(self, body_file):
        return {'body': body_file}


class RequestHeadersDeserializer(ActionDispatcher):
    """Default request headers deserializer."""

    def deserialize(self, request, action):
        return self.dispatch(request, action=action)

    def default(self, request):
        return {}


class RequestDeserializer(object):
    """Break up a Request object into more useful pieces."""

    def __init__(self, body_deserializers=None, headers_deserializer=None):
        self.body_deserializers = {
            'application/json': JSONDeserializer(),
        }
        self.body_deserializers.update(body_deserializers or {})

        self.headers_deserializer = (headers_deserializer or
                                     RequestHeadersDeserializer())

    def deserialize(self, request):
        """Extract necessary pieces of the request.

        :param request: Request object
        :returns: tuple of expected controller action name, dictionary of
                 keyword arguments to pass to the controller, the expected
                 content type of the response

        """
        action_args = self.get_action_args(request.environ)
        action = action_args.pop('action', None)

        action_args.update(self.deserialize_headers(request, action))
        action_args.update(self.deserialize_body(request, action))

        accept = self.get_expected_content_type(request)

        return (action, action_args, accept)

    def deserialize_headers(self, request, action):
        return self.headers_deserializer.deserialize(request, action)

    def deserialize_body(self, request, action):
        try:
            content_type = request.best_match_content_type()
        except exception.InvalidContentType:
            LOG.debug("Unrecognized Content-Type provided in request")
            return {}

        if content_type is None:
            LOG.debug("No Content-Type provided in request")
            return {}

        if not len(request.body) > 0:
            LOG.debug("Empty body provided in request")
            return {}

        try:
            deserializer = self.get_body_deserializer(content_type)
        except exception.InvalidContentType:
            with excutils.save_and_reraise_exception():
                LOG.debug("Unable to deserialize body as provided "
                          "Content-Type")

        if isinstance(deserializer, ZipDeserializer):
            body = request.body_file
        else:
            body = request.body

        return deserializer.deserialize(body, action)

    def get_body_deserializer(self, content_type):
        try:
            return self.body_deserializers[content_type]
        except (KeyError, TypeError):
            raise exception.InvalidContentType(content_type=content_type)

    def get_expected_content_type(self, request):
        return request.best_match_content_type()

    def get_action_args(self, request_environment):
        """Parse dictionary created by routes library."""
        try:
            args = request_environment['wsgiorg.routing_args'][1].copy()
        except Exception:
            return {}

        try:
            del args['controller']
        except KeyError:
            pass

        try:
            del args['format']
        except KeyError:
            pass

        return args


class Application(object):
    """Base WSGI application wrapper. Subclasses need to implement __call__."""

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [app:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [app:wadl]
            latest_version = 1.3
            paste.app_factory = nova.api.fancy_api:Wadl.factory

        which would result in a call to the `Wadl` class as

            import tacker.api.fancy_api
            fancy_api.Wadl(latest_version='1.3')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        return cls(**local_config)

    def __call__(self, environ, start_response):
        r"""Subclasses will probably want to implement __call__ like this:

        @webob.dec.wsgify(RequestClass=Request)
        def __call__(self, req):
          # Any of the following objects work as responses:

          # Option 1: simple string
          res = 'message\n'

          # Option 2: a nicely formatted HTTP exception page
          res = exc.HTTPForbidden(explanation='Nice try')

          # Option 3: a webob Response object (in case you need to play with
          # headers, or you want to be treated like an iterable, or or)
          res = Response();
          res.app_iter = open('somefile')

          # Option 4: any wsgi app to be run next
          res = self.application

          # Option 5: you can get a Response object for a wsgi app, too, to
          # play with headers etc
          res = req.get_response(self.application)

          # You can then just return your response...
          return res
          # ... or set req.response and return None.
          req.response = res

        See the end of http://pythonpaste.org/webob/modules/dec.html
        for more info.

        """
        raise NotImplementedError(_('You must implement __call__'))


class Debug(Middleware):
    """Middleware for debugging.

    Helper class that can be inserted into any WSGI application chain
    to get information about the request and response.
    """

    @webob.dec.wsgify
    def __call__(self, req):
        print(("*" * 40) + " REQUEST ENVIRON")
        for key, value in req.environ.items():
            print(key, "=", value)
        print()
        resp = req.get_response(self.application)

        print(("*" * 40) + " RESPONSE HEADERS")
        for (key, value) in (resp.headers).items():
            print(key, "=", value)
        print()

        resp.app_iter = self.print_generator(resp.app_iter)

        return resp

    @staticmethod
    def print_generator(app_iter):
        """Print contents of a wrapper string iterator when iterated."""
        print(("*" * 40) + " BODY")
        for part in app_iter:
            sys.stdout.write(part)
            sys.stdout.flush()
            yield part
        print()


class DefaultMethodController(object):
    """Controller that handles the OPTIONS request method.

    This controller handles the OPTIONS request method and any of the HTTP
    methods that are not explicitly implemented by the application.
    """

    def options(self, request, **kwargs):
        """Return a response that includes the 'Allow' header.

        Return a response that includes the 'Allow' header listing the methods
        that are implemented. A 204 status code is used for this response.
        """
        headers = [('Allow', kwargs.get('allowed_methods'))]
        raise webob.exc.HTTPNoContent(headers=headers)

    def reject(self, request, **kwargs):
        """Return a 405 method not allowed error.

        As a convenience, the 'Allow' header with the list of implemented
        methods is included in the response as well.
        """
        headers = [('Allow', kwargs.get('allowed_methods'))]
        raise webob.exc.HTTPMethodNotAllowed(
            headers=headers)


class Router(object):
    """WSGI middleware that maps incoming requests to WSGI apps."""

    @classmethod
    def factory(cls, global_config, **local_config):
        """Return an instance of the WSGI Router class."""
        return cls()

    def __init__(self, mapper):
        """Create a router for the given routes.Mapper.

        Each route in `mapper` must specify a 'controller', which is a
        WSGI app to call.  You'll probably want to specify an 'action' as
        well and have your controller be a wsgi.Controller, who will route
        the request to the action method.

        Examples:
          mapper = routes.Mapper()
          sc = ServerController()

          # Explicit mapping of one route to a controller+action
          mapper.connect(None, "/svrlist", controller=sc, action="list")

          # Actions are all implicitly defined
          mapper.resource("network", "networks", controller=nc)

          # Pointing to an arbitrary WSGI app.  You can specify the
          # {path_info:.*} parameter so the target app can be handed just that
          # section of the URL.
          mapper.connect(None, "/v1.0/{path_info:.*}", controller=BlogApp())
        """
        self.map = mapper
        self._setup_routes(self.map)
        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          self.map)

    @webob.dec.wsgify
    def __call__(self, req):
        """Route the incoming request to a controller based on self.map.

        If no match, return a 404.
        """
        return self._router

    @staticmethod
    @webob.dec.wsgify(RequestClass=Request)
    def _dispatch(req):
        """Dispatch a Request.

        Called by self._router after matching the incoming request to a route
        and putting the information into req.environ. Either returns 404
        or the routed WSGI app's response.
        """
        match = req.environ['wsgiorg.routing_args'][1]
        if not match:
            language = req.best_match_language()
            msg = _('The resource could not be found.')
            msg = i18n.translate(msg, language)
            return webob.exc.HTTPNotFound(explanation=msg)
        app = match['controller']
        return app

    def _setup_routes(self, mapper):
        pass


class ResourceExceptionHandler(object):
    """Context manager to handle Resource exceptions.

    Used when processing exceptions generated by API implementation
    methods.  Converts most exceptions to Fault
    exceptions, with the appropriate logging.
    """

    def __enter__(self):
        return None

    def __exit__(self, ex_type, ex_value, ex_traceback):
        if not ex_value:
            return True
        if isinstance(ex_value, exception.Forbidden):
            raise Fault(webob.exc.HTTPForbidden(
                explanation=ex_value.format_message()))
        elif isinstance(ex_value, exception.BadRequest):
            raise Fault(exception.ConvertedException(
                code=ex_value.code,
                explanation=ex_value.format_message()))
        elif isinstance(ex_value, exception.Conflict):
            raise Fault(webob.exc.HTTPConflict(
                explanation=ex_value.format_message()))
        elif isinstance(ex_value, TypeError):
            exc_info = (ex_type, ex_value, ex_traceback)
            LOG.error('Exception handling resource: %s', ex_value,
                      exc_info=exc_info)
            raise Fault(webob.exc.HTTPBadRequest())
        elif isinstance(ex_value, Fault):
            LOG.error("Fault thrown: %s", ex_value)
            raise ex_value
        elif isinstance(ex_value, webob.exc.HTTPException):
            LOG.error("HTTP exception thrown: %s", ex_value)
            raise Fault(ex_value)

        # We didn't handle the exception
        return False


def response(code):
    """Attaches response code to a method.

    This decorator associates a response code with a method.  Note
    that the function attributes are directly manipulated; the method
    is not wrapped.
    """

    def decorator(func):
        func.wsgi_code = code
        return func

    return decorator


class ResponseObject(object):
    """Bundles a response object

    Object that app methods may return in order to allow its response
    to be modified by extensions in the code. Its use is optional (and
    should only be used if you really know what you are doing).
    """

    def __init__(self, obj, code=None, headers=None):
        """Builds a response object."""

        self.obj = obj
        self._default_code = 200
        self._code = code
        self._headers = headers or {}
        self.serializer = JSONDictSerializer()

    def __getitem__(self, key):
        """Retrieves a header with the given name."""

        return self._headers[key.lower()]

    def __setitem__(self, key, value):
        """Sets a header with the given name to the given value."""

        self._headers[key.lower()] = value

    def __delitem__(self, key):
        """Deletes the header with the given name."""

        del self._headers[key.lower()]

    def serialize(self, request, content_type):
        """Serializes the wrapped object.

        Utility method for serializing the wrapped object.  Returns a
        webob.Response object.

        Header values are set to the appropriate Python type and
        encoding demanded by PEP 3333: whatever the native str type is.
        """

        serializer = self.serializer
        if self.obj is None:
            body = None
        elif content_type == 'text/plain':
            body = self.obj
        else:
            body = serializer.serialize(self.obj)
        response = webob.Response(body=body)
        response.status_int = self.code
        for hdr, val in self._headers.items():
            # Headers must be a str that was first safely
            # encoded to UTF-8 (to catch any bad encodings) and then
            # decoded back to a native str.
            response.headers[hdr] = encodeutils.safe_decode(
                encodeutils.safe_encode(val))
        # Deal with content_type
        if not isinstance(content_type, str):
            content_type = str(content_type)

        response.headers['Content-Type'] = encodeutils.safe_decode(
            encodeutils.safe_encode(content_type))
        return response

    @property
    def code(self):
        """Retrieve the response status."""

        return self._code or self._default_code

    @property
    def headers(self):
        """Retrieve the headers."""

        return self._headers.copy()


class Resource(Application):
    """WSGI app that handles (de)serialization and controller dispatch.

    WSGI app that reads routing information supplied by RoutesMiddleware
    and calls the requested action method upon its controller.  All
    controller action methods must accept a 'req' argument, which is the
    incoming wsgi.Request. If the operation is a PUT or POST, the controller
    method must also accept a 'body' argument (the deserialized request body).
    They may raise a webob.exc exception or return a dict, which will be
    serialized by requested content type.

    """

    def __init__(self, controller, deserializer=None, serializer=None):
        """Object initialization.

        :param controller: object that implement methods created by routes lib
        :param deserializer: object that can serialize the output of a
                             controller into a webob response
        :param serializer: object that can deserialize a webob request
                           into necessary pieces
        """
        self.controller = controller
        self.deserializer = deserializer or RequestDeserializer()
        self.serializer = serializer or ResponseSerializer()

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, request):
        """WSGI method that controls (de)serialization and method dispatch."""

        LOG.info("%(method)s %(url)s", {"method": request.method,
                                        "url": request.url})

        try:
            action, args, accept = self.deserializer.deserialize(request)
        except exception.InvalidContentType:
            LOG.exception("InvalidContentType: Unsupported Content-Type")
            return Fault(webob.exc.HTTPUnsupportedMediaType(
                explanation=_("Unsupported Content-Type")))
        except exception.MalformedRequestBody:
            LOG.exception("MalformedRequestBody: Malformed request body")
            return Fault(webob.exc.HTTPBadRequest(
                explanation=_("Malformed request body")))

        response = None
        try:
            with ResourceExceptionHandler():
                action_result = self.dispatch(request, action, args)
        except Fault as ex:
            response = ex
        except Exception:
            raise Fault(webob.exc.HTTPInternalServerError())

        if not response:
            resp_obj = None
            if (isinstance(action_result, (dict, list, str)) or
                    action_result is None):
                resp_obj = ResponseObject(action_result)
            elif isinstance(action_result, ResponseObject):
                resp_obj = action_result
            else:
                response = action_result

            # Run post-processing extensions
            if resp_obj:
                method = getattr(self.controller, action)
                # Do a preserialize to set up the response object
                if hasattr(method, 'wsgi_code'):
                    resp_obj._default_code = method.wsgi_code

            if resp_obj and not response:
                response = resp_obj.serialize(request, accept)

        try:
            msg_dict = dict(url=request.url, status=response.status_int)
            msg = _("%(url)s returned with HTTP %(status)d") % msg_dict
        except AttributeError as e:
            msg_dict = dict(url=request.url, exception=e)
            msg = _("%(url)s returned a fault: %(exception)s") % msg_dict

        LOG.info(msg)

        return response

    def dispatch(self, request, action, action_args):
        """Find action-spefic method on controller and call it."""

        controller_method = getattr(self.controller, action)
        return controller_method(request=request, **action_args)


def _default_body_function(wrapped_exc):
    code = wrapped_exc.status_int
    fault_data = {
        'Error': {
            'code': code,
            'message': wrapped_exc.explanation}}
    # 'code' is an attribute on the fault tag itself
    metadata = {'attributes': {'Error': 'code'}}
    return fault_data, metadata


class Fault(webob.exc.HTTPException):
    """Wrap webob.exc.HTTPException to provide API friendly response."""

    def __init__(self, exception):
        """Create a Fault for the given webob.exc.exception."""
        self.wrapped_exc = exception
        for key, value in list(self.wrapped_exc.headers.items()):
            self.wrapped_exc.headers[key] = str(value)
        self.status_int = exception.status_int

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        """Generate a WSGI response based on the exception passed to ctor."""
        user_locale = req.best_match_language()
        # Replace the body with fault details.
        code = self.wrapped_exc.status_int
        fault_name = http.client.responses[code]
        explanation = self.wrapped_exc.explanation
        LOG.debug("Returning %(code)s to user: %(explanation)s",
                  {'code': code, 'explanation': explanation})

        explanation = i18n.translate(explanation, user_locale)
        fault_data = {}
        if fault_name is not None:
            fault_data['title'] = fault_name
        fault_data['status'] = code
        fault_data['detail'] = explanation

        if code == 413 or code == 429:
            retry = self.wrapped_exc.headers.get('Retry-After', None)
            if retry:
                fault_data['retryAfter'] = retry

        self.wrapped_exc.content_type = 'application/problem+json'
        self.wrapped_exc.charset = 'UTF-8'

        body = JSONDictSerializer().serialize(fault_data)
        if isinstance(body, str):
            body = body.encode('utf-8')
        self.wrapped_exc.body = body

        return self.wrapped_exc

    def __str__(self):
        return self.wrapped_exc.__str__()


# NOTE(salvatore-orlando): this class will go once the
# extension API framework is updated
class Controller(object):
    """WSGI app that dispatched to methods.

    WSGI app that reads routing information supplied by RoutesMiddleware
    and calls the requested action method upon itself.  All action methods
    must, in addition to their normal parameters, accept a 'req' argument
    which is the incoming wsgi.Request.  They raise a webob.exc exception,
    or return a dict which will be serialized by requested content type.

    """

    _view_builder_class = None

    def __init__(self):
        """Initialize controller with a view builder instance."""
        if self._view_builder_class:
            self._view_builder = self._view_builder_class()
        else:
            self._view_builder = None

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        """Call the method specified in req.environ by RoutesMiddleware."""
        arg_dict = req.environ['wsgiorg.routing_args'][1]
        action = arg_dict['action']
        method = getattr(self, action)
        del arg_dict['controller']
        del arg_dict['action']
        if 'format' in arg_dict:
            del arg_dict['format']
        arg_dict['request'] = req
        result = method(**arg_dict)

        if isinstance(result, dict) or result is None:
            if result is None:
                status = 204
                content_type = ''
                body = None
            else:
                status = 200
                content_type = req.best_match_content_type()
                body = self._serialize(result, content_type)

            response = webob.Response(status=status,
                                      content_type=content_type,
                                      body=body)
            msg_dict = dict(url=req.url, status=response.status_int)
            msg = _("%(url)s returned with HTTP %(status)d") % msg_dict
            LOG.debug(msg)
            return response
        else:
            return result

    def _serialize(self, data, content_type):
        """Serialize the given dict to the provided content_type.

        Uses self._serialization_metadata if it exists, which is a dict mapping
        MIME types to information needed to serialize to that type.

        """
        _metadata = getattr(type(self), '_serialization_metadata', {})

        serializer = Serializer(_metadata)
        try:
            return serializer.serialize(data, content_type)
        except exception.InvalidContentType:
            msg = _('The requested content type %s is invalid.') % content_type
            raise webob.exc.HTTPNotAcceptable(msg)

    def _deserialize(self, data, content_type):
        """Deserialize the request body to the specefied content type.

        Uses self._serialization_metadata if it exists, which is a dict mapping
        MIME types to information needed to serialize to that type.

        """
        _metadata = getattr(type(self), '_serialization_metadata', {})
        serializer = Serializer(_metadata)
        return serializer.deserialize(data, content_type)['body']


# NOTE(salvatore-orlando): this class will go once the
# extension API framework is updated
class Serializer(object):
    """Serializes and deserializes dictionaries to certain MIME types."""

    def __init__(self, metadata=None):
        """Create a serializer based on the given WSGI environment.

        'metadata' is an optional dict mapping MIME types to information
        needed to serialize a dictionary to that type.

        """
        self.metadata = metadata or {}

    def _get_serialize_handler(self, content_type):
        handlers = {
            'application/json': JSONDictSerializer(),
        }

        try:
            return handlers[content_type]
        except Exception:
            raise exception.InvalidContentType(content_type=content_type)

    def serialize(self, data, content_type):
        """Serialize a dictionary into the specified content type."""
        return self._get_serialize_handler(content_type).serialize(data)

    def deserialize(self, datastring, content_type):
        """Deserialize a string to a dictionary.

        The string must be in the format of a supported MIME type.

        """
        try:
            return self.get_deserialize_handler(content_type).deserialize(
                datastring)
        except Exception:
            raise webob.exc.HTTPBadRequest(_("Could not deserialize data"))

    def get_deserialize_handler(self, content_type):
        handlers = {
            'application/json': JSONDeserializer(),
        }

        try:
            return handlers[content_type]
        except Exception:
            raise exception.InvalidContentType(content_type=content_type)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add unique constraints on vnfd_id,deleted in vnf_package_vnfd

Revision ID: 975e28392888
Revises: abbef484b34c
Create Date: 2019-12-10 02:40:12.966027

"""

# revision identifiers, used by Alembic.
revision = '975e28392888'
down_revision = 'abbef484b34c'

from alembic import op  # noqa: E402
import sqlalchemy as sa  # noqa: E402
from sqlalchemy.engine import reflection  # noqa: E402


def _migrate_duplicate_vnf_package_vnfd_id(table):

    meta = sa.MetaData(bind=op.get_bind())
    t = sa.Table(table, meta, autoload=True)

    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        dup_vnfd_ids = session.query(t.c.vnfd_id).group_by(
            t.c.vnfd_id).having(sa.func.count() > 1).all()
        if dup_vnfd_ids:
            for vnfd_id in dup_vnfd_ids:
                duplicate_obj_query = session.query(t).filter(
                    t.c.vnfd_id == vnfd_id[0]).all()
                for dup_obj in duplicate_obj_query:
                    if dup_obj.deleted == '1':
                        session.execute(t.update().where(
                            t.c.id == dup_obj.id).values(deleted=dup_obj.id))
    session.commit()

    op.create_unique_constraint(
        constraint_name='uniq_%s0vnfd_id0deleted' % table,
        table_name=table,
        columns=['vnfd_id', 'deleted'])


def upgrade(active_plugins=None, options=None):
    check_constraints = (reflection.Inspector.from_engine(op.get_bind())
                         .get_check_constraints('vnf_package_vnfd'))
    for constraint in check_constraints:
        if '`deleted`' in constraint['sqltext']:
            op.drop_constraint(
                constraint_name=constraint['name'],
                table_name='vnf_package_vnfd',
                type_="check"
            )
            break

    op.alter_column('vnf_package_vnfd',
                    'deleted',
                    type_=sa.String(36), default="0")
    _migrate_duplicate_vnf_package_vnfd_id('vnf_package_vnfd')



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import routes

from oslo_config import cfg
from tacker.api.vnflcm.v1 import controller as vnf_lcm_controller
from tacker import wsgi


class VnflcmAPIRouter(wsgi.Router):
    """Routes requests on the API to the appropriate controller and method."""

    def __init__(self):
        mapper = routes.Mapper()
        super(VnflcmAPIRouter, self).__init__(mapper)

    def _setup_route(self, mapper, url, methods, controller,
                     default_resource):
        all_methods = ['HEAD', 'GET', 'POST', 'PUT', 'PATCH', 'DELETE']
        missing_methods = [m for m in all_methods if m not in methods.keys()]
        allowed_methods_str = ",".join(methods.keys())
        scope_opts = []
        for method, action in methods.items():
            mapper.connect(url,
                           controller=controller,
                           action=action,
                           conditions={'method': [method]})

            add = cfg.ListOpt('vnflcm_' + action + '_scope',
                      default=[],
                      help="OAuth2.0 api token scope for" + action)
            scope_opts.append(add)

        cfg.CONF.register_opts(scope_opts, group='authentication')

        if missing_methods:
            mapper.connect(url,
                           controller=default_resource,
                           action='reject',
                           allowed_methods=allowed_methods_str,
                           conditions={'method': missing_methods})

    def _setup_routes(self, mapper):
        default_resource = wsgi.Resource(wsgi.DefaultMethodController(),
                                         wsgi.RequestDeserializer())

        controller = vnf_lcm_controller.create_resource()

        methods = {"GET": "api_versions"}
        self._setup_route(mapper, "/api_versions",
                          methods, controller, default_resource)

        # Allowed methods on /vnflcm/v1/vnf_instances resource
        methods = {"GET": "index", "POST": "create"}
        self._setup_route(mapper, "/vnf_instances",
                          methods, controller, default_resource)

        # Allowed methods on
        # /vnflcm/v1/vnf_instances/{vnfInstanceId} resource
        methods = {"DELETE": "delete", "GET": "show", "PATCH": "update"}
        self._setup_route(mapper, "/vnf_instances/{id}",
                          methods, controller, default_resource)

        # Allowed methods on
        # /vnflcm/v1/vnf_instances/{vnfInstanceId}/instantiate resource
        methods = {"POST": "instantiate"}
        self._setup_route(mapper,
                "/vnf_instances/{id}/instantiate",
                methods, controller, default_resource)

        # Allowed methods on
        # /vnflcm/v1/vnf_instances/{vnfInstanceId}/heal resource
        methods = {"POST": "heal"}
        self._setup_route(mapper,
                "/vnf_instances/{id}/heal",
                methods, controller, default_resource)

        # Allowed methods on
        # /vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId} resource
        methods = {"GET": "show_lcm_op_occs"}
        self._setup_route(mapper, "/vnf_lcm_op_occs/{id}",
                          methods, controller, default_resource)

        # Allowed methods on
        # /vnflcm/v1/vnf_instances/{vnfInstanceId}/terminate resource
        methods = {"POST": "terminate"}
        self._setup_route(mapper,
                "/vnf_instances/{id}/terminate",
                methods, controller, default_resource)

        # Allowed methods on
        # /vnflcm/v1/vnf_instances/{vnfInstanceId}/heal resource
        methods = {"POST": "heal"}
        self._setup_route(mapper,
                "/vnf_instances/{id}/heal",
                methods, controller, default_resource)

        # Allowed methods on
        # /vnflcm/v1/vnf_instances/{vnfInstanceId}/scale resource
        methods = {"POST": "scale"}
        self._setup_route(mapper,
                "/vnf_instances/{id}/scale",
                methods, controller, default_resource)

        # Allowed methods on
        # {apiRoot}/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/rollback resource
        methods = {"POST": "rollback"}
        self._setup_route(mapper,
                "/vnf_lcm_op_occs/{id}/rollback",
                methods, controller, default_resource)

        # {apiRoot}/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/cancel resource
        methods = {"POST": "cancel"}
        self._setup_route(mapper,
                          "/vnf_lcm_op_occs/{id}/cancel",
                          methods, controller, default_resource)

        # {apiRoot}/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/fail resource
        methods = {"POST": "fail"}
        self._setup_route(mapper,
                "/vnf_lcm_op_occs/{id}/fail",
                methods, controller, default_resource)

        # {apiRoot}/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/retry resource
        methods = {"POST": "retry"}
        self._setup_route(mapper,
                "/vnf_lcm_op_occs/{id}/retry",
                methods, controller, default_resource)

        methods = {"GET": "subscription_list", "POST": "register_subscription"}
        self._setup_route(mapper, "/subscriptions",
                methods, controller, default_resource)

        methods = {"GET": "subscription_show", "DELETE": "delete_subscription"}
        self._setup_route(mapper, "/subscriptions/{subscriptionId}",
                methods, controller, default_resource)

        # {apiRoot}/vnflcm/v1/vnf_lcm_op_occs/{vnfLcmOpOccId}/retry resource
        methods = {"POST": "retry"}
        self._setup_route(mapper,
                "/vnf_lcm_op_occs/{id}/retry",
                methods, controller, default_resource)

        # Allowed methods on
        # {apiRoot}/vnflcm/v1/vnf_lcm_op_occs resource
        methods = {"GET": "list_lcm_op_occs"}
        self._setup_route(mapper, "/vnf_lcm_op_occs",
                          methods, controller, default_resource)

        # {apiRoot}/vnf_instances/{vnfInstanceId}/change_ext_conn resource
        methods = {"POST": "change_ext_conn"}
        self._setup_route(mapper, "/vnf_instances/{id}/change_ext_conn",
                methods, controller, default_resource)



# Copyright (c) 2020 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import ddt
from oslo_config import cfg

from tacker.tests.unit import base
from tacker.tests.unit.vnflcm import fakes
from tacker.tests import uuidsentinel
from tacker.vnflcm import utils as vnflcm_utils


@ddt.ddt
class VnfLcmUtilsTestCase(base.TestCase):

    @ddt.data(
        {'image_path': 'cirros-0.5.2-x86_64-disk.img',
         'extracted_path': 'cirros-0.5.2-x86_64-disk.img'},
        {'image_path': '../ImageFiles/image/cirros-0.5.2-x86_64-disk.img',
         'extracted_path': 'ImageFiles/image/cirros-0.5.2-x86_64-disk.img'},
        {'image_path': '../../Files/image/cirros-0.5.2-x86_64-disk.img',
         'extracted_path': 'Files/image/cirros-0.5.2-x86_64-disk.img'}
    )
    @ddt.unpack
    def test_create_grant_request_with_software_image_path(self, image_path,
                                                           extracted_path):
        vnf_package_id = uuidsentinel.package_uuid
        vnfd_dict = fakes.get_vnfd_dict(image_path=image_path)
        vnf_software_images = vnflcm_utils._create_grant_request(
            vnfd_dict, vnf_package_id)
        vnf_package_path = cfg.CONF.vnf_package.vnf_package_csar_path
        expected_image_path = os.path.join(vnf_package_path, vnf_package_id,
                                           extracted_path)
        self.assertEqual(expected_image_path,
                         vnf_software_images['VDU1'].image_path)

    def test_get_param_data_with_flavour_description(self):
        vnfd_dict = fakes.get_vnfd_dict()
        vnfd_dict.update({'imports': []})
        instantiate_vnf_req = fakes.get_instantiate_vnf_request_obj()
        param_value = vnflcm_utils._get_param_data(vnfd_dict,
                                                   instantiate_vnf_req)
        expected_flavour_description = 'A simple flavor'
        self.assertEqual(expected_flavour_description,
                         param_value['flavour_description'])

    def test_topology_template_param_of_vnf_dict(self):
        vnf_dict = fakes.vnf_dict()
        vnf_keys = vnf_dict['vnfd']['attributes']['vnfd_simple']
        self.assertIn('node_templates', vnf_keys)
        self.assertIn('policies', vnf_keys)
        self.assertIn('groups', vnf_keys)



# (c) Copyright 2014 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from tacker.api import extensions
from tacker.api.v1 import base
from tacker import manager
from tacker.plugins.common import constants


def build_plural_mappings(special_mappings, resource_map):
    """Create plural to singular mapping for all resources.

    Allows for special mappings to be provided, like policies -> policy.
    Otherwise, will strip off the last character for normal mappings, like
    routers -> router.
    """
    plural_mappings = {}
    for plural in resource_map:
        singular = special_mappings.get(plural)
        if not singular:
            if plural.endswith('ies'):
                singular = "%sy" % plural[:-3]
            else:
                singular = plural[:-1]
        plural_mappings[plural] = singular
    return plural_mappings


def build_resource_info(plural_mappings, resource_map, which_service,
                        action_map=None,
                        translate_name=False, allow_bulk=False):
    """Build resources for advanced services.

    Takes the resource information, and singular/plural mappings, and creates
    API resource objects for advanced services extensions. Will optionally
    translate underscores to dashes in resource names, register the resource,
    and accept action information for resources.

    :param plural_mappings: mappings between singular and plural forms
    :param resource_map: attribute map for the WSGI resources to create
    :param which_service: The name of the service for which the WSGI resources
                          are being created. This name will be used to pass
                          the appropriate plugin to the WSGI resource.
                          It can be set to None or "CORE"to create WSGI
                          resources for the core plugin
    :param action_map: custom resource actions
    :param translate_name: replaces underscores with dashes
    :param allow_bulk: True if bulk create are allowed
    """
    resources = []
    if not which_service:
        which_service = constants.CORE
    action_map = action_map or {}
    plugin = manager.TackerManager.get_service_plugins()[which_service]
    for collection_name in resource_map:
        resource_name = plural_mappings[collection_name]
        params = resource_map.get(collection_name, {})
        if translate_name:
            collection_name = collection_name.replace('_', '-')
        member_actions = action_map.get(resource_name, {})
        controller = base.create_resource(
            collection_name, resource_name, plugin, params,
            member_actions=member_actions,
            allow_bulk=allow_bulk,
            allow_pagination=cfg.CONF.allow_pagination,
            allow_sorting=cfg.CONF.allow_sorting)
        resource = extensions.ResourceExtension(
            collection_name,
            controller,
            path_prefix=constants.COMMON_PREFIXES[which_service],
            member_actions=member_actions,
            attr_map=params)
        resources.append(resource)
    return resources



# Copyright 2011 OpenStack Foundation.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import os

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
import routes
import webob.dec
import webob.exc

from tacker._i18n import _
from tacker.common import exceptions
import tacker.extensions
from tacker import policy
from tacker import wsgi


LOG = logging.getLogger(__name__)


class PluginInterface(object, metaclass=abc.ABCMeta):

    @classmethod
    def __subclasshook__(cls, klass):
        """Checking plugin class.

        The __subclasshook__ method is a class method
        that will be called every time a class is tested
        using issubclass(klass, PluginInterface).
        In that case, it will check that every method
        marked with the abstractmethod decorator is
        provided by the plugin class.
        """

        if not cls.__abstractmethods__:
            return NotImplemented

        for method in cls.__abstractmethods__:
            if any(method in base.__dict__ for base in klass.__mro__):
                continue
            return NotImplemented
        return True


class ExtensionDescriptor(object):
    """Base class that defines the contract for extensions.

    Note that you don't have to derive from this class to have a valid
    extension; it is purely a convenience.
    """

    def get_name(self):
        """The name of the extension.

        e.g. 'Fox In Socks'
        """
        raise NotImplementedError()

    def get_alias(self):
        """The alias for the extension.

        e.g. 'FOXNSOX'
        """
        raise NotImplementedError()

    def get_description(self):
        """Friendly description for the extension.

        e.g. 'The Fox In Socks Extension'
        """
        raise NotImplementedError()

    def get_namespace(self):
        """The XML namespace for the extension.

        e.g. 'http://www.fox.in.socks/api/ext/pie/v1.0'
        """
        raise NotImplementedError()

    def get_updated(self):
        """The timestamp when the extension was last updated.

        e.g. '2011-01-22T13:25:27-06:00'
        """
        # NOTE(justinsb): Not sure of the purpose of this is, vs the XML NS
        raise NotImplementedError()

    def get_resources(self):
        """List of extensions.ResourceExtension extension objects.

        Resources define new nouns, and are accessible through URLs.
        """
        resources = []
        return resources

    def get_actions(self):
        """List of extensions.ActionExtension extension objects.

        Actions are verbs callable from the API.
        """
        actions = []
        return actions

    def get_request_extensions(self):
        """List of extensions.RequestException extension objects.

        Request extensions are used to handle custom request data.
        """
        request_exts = []
        return request_exts

    def get_extended_resources(self, version):
        """Retrieve extended resources or attributes for core resources.

        Extended attributes are implemented by a core plugin similarly
        to the attributes defined in the core, and can appear in
        request and response messages. Their names are scoped with the
        extension's prefix. The core API version is passed to this
        function, which must return a
        map[<resource_name>][<attribute_name>][<attribute_property>]
        specifying the extended resource attribute properties required
        by that API version.

        Extension can add resources and their attr definitions too.
        The returned map can be integrated into RESOURCE_ATTRIBUTE_MAP.
        """
        return {}

    def get_plugin_interface(self):
        """Returns an abstract class which defines contract for the plugin.

        The abstract class should inherit from extesnions.PluginInterface,
        Methods in this abstract class  should be decorated as abstractmethod
        """
        return None

    def update_attributes_map(self, extended_attributes,
                              extension_attrs_map=None):
        """Update attributes map for this extension.

        This is default method for extending an extension's attributes map.
        An extension can use this method and supplying its own resource
        attribute map in extension_attrs_map argument to extend all its
        attributes that needs to be extended.

        If an extension does not implement update_attributes_map, the method
        does nothing and just return.
        """
        if not extension_attrs_map:
            return

        for resource, attrs in extension_attrs_map.items():
            extended_attrs = extended_attributes.get(resource)
            if extended_attrs:
                attrs.update(extended_attrs)

    def get_alias_namespace_compatibility_map(self):
        """Returns mappings between extension aliases and XML namespaces.

        The mappings are XML namespaces that should, for backward compatibility
        reasons, be added to the XML serialization of extended attributes.
        This allows an established extended attribute to be provided by
        another extension than the original one while keeping its old alias
        in the name.
        :return: A dictionary of extension_aliases and namespace strings.
        """
        return {}


class ActionExtensionController(wsgi.Controller):

    def __init__(self, application):
        self.application = application
        self.action_handlers = {}

    def add_action(self, action_name, handler):
        self.action_handlers[action_name] = handler

    def action(self, request, id):
        input_dict = self._deserialize(request.body,
                                       request.get_content_type())
        for action_name, handler in (self.action_handlers).items():
            if action_name in input_dict:
                return handler(input_dict, request, id)
        # no action handler found (bump to downstream application)
        response = self.application
        return response


class RequestExtensionController(wsgi.Controller):

    def __init__(self, application):
        self.application = application
        self.handlers = []

    def add_handler(self, handler):
        self.handlers.append(handler)

    def process(self, request, *args, **kwargs):
        res = request.get_response(self.application)
        # currently request handlers are un-ordered
        for handler in self.handlers:
            response = handler(request, res)
        return response


class ExtensionController(wsgi.Controller):

    def __init__(self, extension_manager):
        self.extension_manager = extension_manager

    def _translate(self, ext):
        ext_data = {}
        ext_data['name'] = ext.get_name()
        ext_data['alias'] = ext.get_alias()
        ext_data['description'] = ext.get_description()
        ext_data['namespace'] = ext.get_namespace()
        ext_data['updated'] = ext.get_updated()
        ext_data['links'] = []  # TODO(dprince): implement extension links
        return ext_data

    def index(self, request):
        extensions = []
        for _alias, ext in (self.extension_manager.extensions).items():
            extensions.append(self._translate(ext))
        return dict(extensions=extensions)

    def show(self, request, id):
        # NOTE(dprince): the extensions alias is used as the 'id' for show
        ext = self.extension_manager.extensions.get(id)
        if not ext:
            raise webob.exc.HTTPNotFound(
                _("Extension with alias %s does not exist") % id)
        return dict(extension=self._translate(ext))

    def delete(self, request, id):
        msg = _('Resource not found.')
        raise webob.exc.HTTPNotFound(msg)

    def create(self, request):
        msg = _('Resource not found.')
        raise webob.exc.HTTPNotFound(msg)


class ExtensionMiddleware(wsgi.Middleware):
    """Extensions middleware for WSGI."""

    def __init__(self, application,
                 ext_mgr=None):
        self.ext_mgr = (ext_mgr
                        or ExtensionManager(get_extensions_path()))
        mapper = routes.Mapper()

        # extended resources
        for resource in self.ext_mgr.get_resources():
            path_prefix = resource.path_prefix
            if resource.parent:
                path_prefix = (resource.path_prefix +
                               "/%s/{%s_id}" %
                               (resource.parent["collection_name"],
                                resource.parent["member_name"]))

            LOG.debug('Extended resource: %s', resource.collection)
            for action, method in (resource.collection_actions).items():
                conditions = dict(method=[method])
                path = "/%s/%s" % (resource.collection, action)
                with mapper.submapper(controller=resource.controller,
                                      action=action,
                                      path_prefix=path_prefix,
                                      conditions=conditions) as submap:
                    submap.connect(path)
                    submap.connect("%s.:(format)" % path)

            mapper.resource(resource.collection, resource.collection,
                            controller=resource.controller,
                            member=resource.member_actions,
                            parent_resource=resource.parent,
                            path_prefix=path_prefix)

        # extended actions
        action_controllers = self._action_ext_controllers(application,
                                                          self.ext_mgr, mapper)
        for action in self.ext_mgr.get_actions():
            LOG.debug('Extended action: %s', action.action_name)
            controller = action_controllers[action.collection]
            controller.add_action(action.action_name, action.handler)

        # extended requests
        req_controllers = self._request_ext_controllers(application,
                                                        self.ext_mgr, mapper)
        for request_ext in self.ext_mgr.get_request_extensions():
            LOG.debug('Extended request: %s', request_ext.key)
            controller = req_controllers[request_ext.key]
            controller.add_handler(request_ext.handler)

        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          mapper)
        super(ExtensionMiddleware, self).__init__(application)

    @classmethod
    def factory(cls, global_config, **local_config):
        """Paste factory."""
        def _factory(app):
            return cls(app, global_config, **local_config)
        return _factory

    def _action_ext_controllers(self, application, ext_mgr, mapper):
        """Return a dict of ActionExtensionController-s by collection."""
        action_controllers = {}
        for action in ext_mgr.get_actions():
            if action.collection not in action_controllers:
                controller = ActionExtensionController(application)
                mapper.connect("/%s/:(id)/action.:(format)" %
                               action.collection,
                               action='action',
                               controller=controller,
                               conditions=dict(method=['POST']))
                mapper.connect("/%s/:(id)/action" % action.collection,
                               action='action',
                               controller=controller,
                               conditions=dict(method=['POST']))
                action_controllers[action.collection] = controller

        return action_controllers

    def _request_ext_controllers(self, application, ext_mgr, mapper):
        """Returns a dict of RequestExtensionController-s by collection."""
        request_ext_controllers = {}
        for req_ext in ext_mgr.get_request_extensions():
            if req_ext.key not in request_ext_controllers:
                controller = RequestExtensionController(application)
                mapper.connect(req_ext.url_route + '.:(format)',
                               action='process',
                               controller=controller,
                               conditions=req_ext.conditions)

                mapper.connect(req_ext.url_route,
                               action='process',
                               controller=controller,
                               conditions=req_ext.conditions)
                request_ext_controllers[req_ext.key] = controller

        return request_ext_controllers

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        """Route the incoming request with router."""
        req.environ['extended.app'] = self.application
        return self._router

    @staticmethod
    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def _dispatch(req):
        """Dispatch the request.

        Returns the routed WSGI app's response or defers to the extended
        application.
        """
        match = req.environ['wsgiorg.routing_args'][1]
        if not match:
            return req.environ['extended.app']
        app = match['controller']
        return app


def extension_middleware_factory(global_config, **local_config):
    """Paste factory."""
    def _factory(app):
        ext_mgr = ExtensionManager.get_instance()
        return ExtensionMiddleware(app, ext_mgr=ext_mgr)
    return _factory


class ExtensionManager(object):
    """Load extensions from the configured extension path.

    See tests/unit/extensions/foxinsocks.py for an
    example extension implementation.
    """

    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls(get_extensions_path())
        return cls._instance

    def __init__(self, path):
        LOG.info('Initializing extension manager.')
        self.path = path
        self.extensions = {}
        self._load_all_extensions()
        policy.reset()

    def get_resources(self):
        """Returns a list of ResourceExtension objects."""
        resources = []
        resources.append(ResourceExtension('extensions',
                                           ExtensionController(self)))
        for ext in self.extensions.values():
            try:
                resources.extend(ext.get_resources())
            except AttributeError:
                # NOTE(dprince): Extension aren't required to have resource
                # extensions
                pass
        return resources

    def get_actions(self):
        """Returns a list of ActionExtension objects."""
        actions = []
        for ext in self.extensions.values():
            try:
                actions.extend(ext.get_actions())
            except AttributeError:
                # NOTE(dprince): Extension aren't required to have action
                # extensions
                pass
        return actions

    def get_request_extensions(self):
        """Returns a list of RequestExtension objects."""
        request_exts = []
        for ext in self.extensions.values():
            try:
                request_exts.extend(ext.get_request_extensions())
            except AttributeError:
                # NOTE(dprince): Extension aren't required to have request
                # extensions
                pass
        return request_exts

    def extend_resources(self, version, attr_map):
        """Extend resources with additional resources or attributes.

        :param attr_map: the existing mapping from resource name to
                         attrs definition.

        After this function, we will extend the attr_map if an extension
        wants to extend this map.
        """
        update_exts = []
        processed_exts = set()
        exts_to_process = self.extensions.copy()
        # Iterate until there are unprocessed extensions or if no progress
        # is made in a whole iteration
        while exts_to_process:
            processed_ext_count = len(processed_exts)
            for ext_name, ext in list(exts_to_process.items()):
                if not hasattr(ext, 'get_extended_resources'):
                    del exts_to_process[ext_name]
                    continue
                if hasattr(ext, 'update_attributes_map'):
                    update_exts.append(ext)
                if hasattr(ext, 'get_required_extensions'):
                    # Process extension only if all required extensions
                    # have been processed already
                    required_exts_set = set(ext.get_required_extensions())
                    if required_exts_set - processed_exts:
                        continue
                try:
                    extended_attrs = ext.get_extended_resources(version)
                    for resource, resource_attrs in extended_attrs.items():
                        if attr_map.get(resource):
                            attr_map[resource].update(resource_attrs)
                        else:
                            attr_map[resource] = resource_attrs
                except AttributeError:
                    LOG.exception("Error fetching extended attributes for "
                                  "extension '%s'", ext.get_name())
                processed_exts.add(ext_name)
                del exts_to_process[ext_name]
            if len(processed_exts) == processed_ext_count:
                # Exit loop as no progress was made
                break
        if exts_to_process:
            # NOTE(salv-orlando): Consider whether this error should be fatal
            LOG.error("It was impossible to process the following "
                      "extensions: %s because of missing requirements.",
                      ','.join(exts_to_process.keys()))

        # Extending extensions' attributes map.
        for ext in update_exts:
            ext.update_attributes_map(attr_map)

    def _check_extension(self, extension):
        """Checks for required methods in extension objects."""
        try:
            LOG.debug('Ext name: %s', extension.get_name())
            LOG.debug('Ext alias: %s', extension.get_alias())
            LOG.debug('Ext description: %s', extension.get_description())
            LOG.debug('Ext namespace: %s', extension.get_namespace())
            LOG.debug('Ext updated: %s', extension.get_updated())
        except AttributeError as ex:
            LOG.exception("Exception loading extension: %s", ex)
            return False
        return True

    def _load_all_extensions(self):
        """Load extensions from the configured path.

        The extension name is constructed from the module_name. If your
        extension module is named widgets.py, the extension class within that
        module should be 'Widgets'.

        See tests/unit/extensions/foxinsocks.py for an example extension
        implementation.
        """
        for path in self.path.split(':'):
            if os.path.exists(path):
                self._load_all_extensions_from_path(path)
            else:
                LOG.error("Extension path '%s' doesn't exist!", path)

    def _load_all_extensions_from_path(self, path):
        # Sorting the extension list makes the order in which they
        # are loaded predictable across a cluster of load-balanced
        # Tacker Servers
        for f in sorted(os.listdir(path)):
            try:
                LOG.debug('Loading extension file: %s', f)
                mod_name, file_ext = os.path.splitext(os.path.split(f)[-1])
                ext_path = os.path.join(path, f)
                if file_ext.lower() == '.py' and not mod_name.startswith('_'):
                    mod = importutils.import_module(
                        'tacker.extensions.' + mod_name)
                    ext_name = mod_name[0].upper() + mod_name[1:]
                    new_ext_class = getattr(mod, ext_name, None)
                    if not new_ext_class:
                        LOG.warning('Did not find expected name '
                                    '"%(ext_name)s" in %(file)s',
                                    {'ext_name': ext_name,
                                     'file': ext_path})
                        continue
                    new_ext = new_ext_class()
                    self.add_extension(new_ext)
            except Exception as exception:
                LOG.warning("Extension file %(f)s wasn't loaded due to "
                            "%(exception)s",
                            {'f': f, 'exception': exception})

    def add_extension(self, ext):
        # Do nothing if the extension doesn't check out
        if not self._check_extension(ext):
            return

        alias = ext.get_alias()
        LOG.info('Loaded extension: %s', alias)

        if alias in self.extensions:
            raise exceptions.DuplicatedExtension(alias=alias)
        self.extensions[alias] = ext


class RequestExtension(object):
    """Extend requests and responses of core Tacker OpenStack API controllers.

    Provide a way to add data to responses and handle custom request data
    that is sent to core Tacker OpenStack API controllers.
    """

    def __init__(self, method, url_route, handler):
        self.url_route = url_route
        self.handler = handler
        self.conditions = dict(method=[method])
        self.key = "%s-%s" % (method, url_route)


class ActionExtension(object):
    """Add custom actions to core Tacker OpenStack API controllers."""

    def __init__(self, collection, action_name, handler):
        self.collection = collection
        self.action_name = action_name
        self.handler = handler


class ResourceExtension(object):
    """Add top level resources to the OpenStack API in Tacker."""

    def __init__(self, collection, controller, parent=None, path_prefix="",
                 collection_actions={}, member_actions={}, attr_map={}):
        self.collection = collection
        self.controller = controller
        self.parent = parent
        self.collection_actions = collection_actions
        self.member_actions = member_actions
        self.path_prefix = path_prefix
        self.attr_map = attr_map


# Returns the extension paths from a config entry and the __path__
# of tacker.extensions
def get_extensions_path():
    paths = ':'.join(tacker.extensions.__path__)
    if cfg.CONF.api_extensions_path:
        paths = ':'.join([cfg.CONF.api_extensions_path, paths])

    return paths


def append_api_extensions_path(paths):
    paths = [cfg.CONF.api_extensions_path] + paths
    cfg.CONF.set_override('api_extensions_path',
                          ':'.join([p for p in paths if p]))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import base as ovoo_base
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.vnfm import vnfm_db
from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


def _vnf_update(context, values):
    update = {'status': values.status,
              'updated_at': values.updated_at}

    api.model_query(context, vnfm_db.VNF). \
        filter_by(id=values.id). \
        update(update, synchronize_session=False)


@db_api.context_manager.reader
def _vnf_get(context, id, columns_to_join=None):
    vnf_data = api.model_query(
        context,
        vnfm_db.VNF).filter_by(
        id=id).filter_by(
            deleted_at='0001-01-01 00:00:00').first()
    if vnf_data:
        vnf_data = vnf_data.__dict__
        vnf_attribute_data = api.model_query(
            context, vnfm_db.VNFAttribute).filter_by(
            vnf_id=vnf_data.get('id')).first()
        vnf_data['vnf_attribute'] = vnf_attribute_data.__dict__
        vnfd_data = api.model_query(
            context, vnfm_db.VNFD).filter_by(
            id=vnf_data.get('vnfd_id')).first()
        vnf_data['vnfd'] = vnfd_data.__dict__
        vnfd_attribute_data = api.model_query(
            context, vnfm_db.VNFDAttribute).filter_by(
            vnfd_id=vnf_data.get('vnfd_id')).first()
        vnf_data['vnfd_attribute'] = vnfd_attribute_data.__dict__
    else:
        vnf_data = ""

    return vnf_data


@base.TackerObjectRegistry.register
class VNF(base.TackerObject, base.TackerObjectDictCompat,
          base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'tenant_id': fields.UUIDField(nullable=False),
        'name': fields.StringField(nullable=False),
        'vnfd_id': fields.UUIDField(nullable=False),
        'instance_id': fields.StringField(nullable=True),
        'mgmt_ip_address': fields.StringField(nullable=True),
        'status': fields.StringField(nullable=True),
        'description': fields.StringField(nullable=True),
        'placement_attr': fields.StringField(nullable=True),
        'vim_id': fields.StringField(nullable=False),
        'error_reason': fields.StringField(nullable=True),
        'vnf_attribute': fields.ObjectField(
            'VNFAttribute', nullable=True),
        'vnfd': fields.ObjectField('VNFD', nullable=True),
    }

    @base.remotable
    def save(self):
        updates = self.obj_clone()
        _vnf_update(self._context, updates)

    @base.remotable_classmethod
    def vnf_index_list(cls, id, context):
        # get vnf_instance data
        expected_attrs = ["vnf_attribute", "vnfd"]
        db_vnf = _vnf_get(context, id, columns_to_join=expected_attrs)
        return db_vnf


@base.TackerObjectRegistry.register
class VnfList(ovoo_base.ObjectListBase, base.TackerObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('VNF')
    }

    @base.remotable_classmethod
    def vnf_index_list(cls, id, context):
        # get vnf_instance data
        expected_attrs = ["vnf_attribute", "vnfd"]
        db_vnf = _vnf_get(context, id, columns_to_join=expected_attrs)
        return db_vnf
